import pytest

from datarobot import (BacktestSpecification, DatetimeModel, DatetimePartitioning,
                       DatetimePartitioningSpecification, FeatureSettings, Periodicity,
                       Project, enums)
from datarobot.utils import parse_time


@pytest.fixture
def holdout_start_date():
    return parse_time('2015-12-10T19:00:00.000000Z')


@pytest.fixture
def datetime_partition_without_holdout_server_data(project_id):
    return {
        'primaryTrainingEndDate': None,
        'holdoutRowCount': 0,
        'backtests': [
            {
                'index': 0,
                'validationRowCount': 108,
                'primaryTrainingDuration': 'P0Y0M48DT0H5M36S',
                'primaryTrainingEndDate': '2014-01-07T22:44:23.000000Z',
                'availableTrainingStartDate': '2013-11-20T22:38:47.000000Z',
                'primaryTrainingStartDate': '2013-11-20T22:38:47.000000Z',
                'validationEndDate': '2014-01-12T22:44:23.000000Z',
                'availableTrainingDuration': 'P0Y0M48DT0H5M36S',
                'availableTrainingRowCount': 4168,
                'gapEndDate': '2014-01-07T22:44:23.000000Z',
                'validationDuration': 'P5D',
                'gapStartDate': '2014-01-07T22:44:23.000000Z',
                'availableTrainingEndDate': '2014-01-07T22:44:23.000000Z',
                'primaryTrainingRowCount': 4168,
                'validationStartDate': '2014-01-07T22:44:23.000000Z',
                'totalRowCount': 4276,
                'gapRowCount': 0,
                'gapDuration': 'P0Y0M0D'
            }
        ],
        'availableTrainingDuration': 'P0Y0M53DT0H5M36S',
        'availableTrainingRowCount': 4276,
        'gapEndDate': None,
        'partitioningWarnings': [],
        'primaryTrainingRowCount': 0,
        'holdoutStartDate': None,
        'datetimePartitionColumn': 'Timestamp',
        'dateFormat': '%Y-%m-%d %H:%M:%S',
        'useTimeSeries': False,
        'defaultToAPriori': False,
        'defaultToKnownInAdvance': False,
        'featureDerivationWindowStart': None,
        'featureDerivationWindowEnd': None,
        'forecastWindowStart': None,
        'forecastWindowEnd': None,
        'numberOfBacktests': 1,
        'availableTrainingStartDate': '2013-11-20T22:38:47.000000Z',
        'primaryTrainingStartDate': None,
        'holdoutEndDate': None,
        'gapStartDate': None,
        'totalRowCount': 4276,
        'gapRowCount': 0,
        'primaryTrainingDuration': 'P0Y0M0D',
        'holdoutDuration': 'P0Y0M0D',
        'gapDuration': 'P0Y0M0D',
        'autopilotDataSelectionMethod': 'duration',
        'validationDuration': 'P5D',
        'projectId': project_id,
        'availableTrainingEndDate': '2014-01-12T22:44:23.000000Z'
    }


@pytest.fixture
def datetime_partition_server_data(project_id, holdout_start_date):
    return {
        'primaryTrainingEndDate': '2015-12-10T19:00:00.000000Z',
        'primaryTrainingDuration': 'P0Y1M0D',
        'backtests': [
            {
                'primaryTrainingEndDate': '2015-11-26T19:00:00.000000Z',
                'primaryTrainingDuration': 'P0Y1M0D',
                'index': 0,
                'availableTrainingStartDate': '2015-06-08T20:00:00.000000Z',
                'primaryTrainingStartDate': '2015-10-26T19:00:00.000000Z',
                'validationEndDate': '2015-12-10T19:00:00.000000Z',
                'availableTrainingDuration': 'P0Y0M170DT23H0M0S',
                'validationDuration': 'P14D',
                'gapStartDate': '2015-11-26T19:00:00.000000Z',
                'availableTrainingEndDate': '2015-11-26T19:00:00.000000Z',
                'validationStartDate': '2015-11-26T19:00:00.000000Z',
                'gapEndDate': '2015-11-26T19:00:00.000000Z',
                'gapDuration': 'P0Y0M0D'
            }
        ],
        'datetimePartitionColumn': 'dates',
        'useTimeSeries': False,
        'defaultToAPriori': False,
        'defaultToKnownInAdvance': False,
        'featureDerivationWindowStart': None,
        'featureDerivationWindowEnd': None,
        'forecastWindowStart': None,
        'forecastWindowEnd': None,
        'dateFormat': '%Y-%m-%d %H:%M:%S',
        'projectId': project_id,
        'availableTrainingStartDate': '2015-06-08T20:00:00.000000Z',
        'primaryTrainingStartDate': '2015-11-10T19:00:00.000000Z',
        'holdoutEndDate': '2015-12-24T19:00:00.000000Z',
        'validationDuration': 'P14D',
        'availableTrainingDuration': 'P0Y0M184DT23H0M0S',
        'numberOfBacktests': 1,
        'gapStartDate': '2015-12-10T19:00:00.000000Z',
        'availableTrainingEndDate': '2015-12-10T19:00:00.000000Z',
        'holdoutStartDate': holdout_start_date.isoformat(),
        'gapEndDate': '2015-12-10T19:00:00.000000Z',
        'holdoutDuration': 'P14D',
        'gapDuration': 'P0Y0M0D',
        'autopilotDataSelectionMethod': 'rowCount'
    }


@pytest.fixture
def datetime_partition_server_data_feature_settings(datetime_partition_server_data):
    settings = dict(datetime_partition_server_data)
    settings.update({
        'featureSettings': [
            {'featureName': 'timestamp', 'knownInAdvance': True, 'aPriori': True},
            {'featureName': 'target', 'knownInAdvance': True, 'aPriori': True},
            {'featureName': 'series_id', 'knownInAdvance': False, 'aPriori': False},
            {'featureName': 'input3', 'knownInAdvance': False, 'aPriori': False},
            {'featureName': 'input4', 'knownInAdvance': True, 'aPriori': True}
        ],
    })
    return settings


@pytest.fixture
def datetime_partition_server_data_extra_time_series_controls(datetime_partition_server_data):
    settings = dict(datetime_partition_server_data)
    settings.update({
        'periodicities': [
            {'timeSteps': 10, 'timeUnit': 'HOUR'},
            {'timeSteps': 600, 'timeUnit': 'MINUTE'},
            {'timeSteps': 7, 'timeUnit': 'DAY'},
        ],
        'treatAsExponential': enums.TREAT_AS_EXPONENTIAL.ALWAYS,
        'differencingMethod': enums.DIFFERENCING_METHOD.SIMPLE
    })
    return settings


@pytest.fixture
def datetime_partition_time_series_server_data(datetime_partition_server_data):
    settings = dict(datetime_partition_server_data)
    time_series_settings = {'useTimeSeries': True, 'defaultToKnownInAdvance': True,
                            'featureDerivationWindowStart': -9, 'featureDerivationWindowEnd': -5,
                            'forecastWindowStart': 5, 'forecastWindowEnd': 9}
    settings.update(time_series_settings)
    return settings


@pytest.fixture
def datetime_partition_after_target_server_data(project_id, holdout_start_date):
    return {
        'primaryTrainingEndDate': '2015-12-10T19:00:00.000000Z',
        'primaryTrainingDuration': 'P0Y1M0D',
        'primaryTrainingRowCount': 100,
        'backtests': [
            {
                'primaryTrainingEndDate': '2015-11-26T19:00:00.000000Z',
                'primaryTrainingDuration': 'P0Y1M0D',
                'primaryTrainingRowCount': 100,
                'index': 0,
                'availableTrainingStartDate': '2015-06-08T20:00:00.000000Z',
                'primaryTrainingStartDate': '2015-10-26T19:00:00.000000Z',
                'validationEndDate': '2015-12-10T19:00:00.000000Z',
                'availableTrainingDuration': 'P0Y0M170DT23H0M0S',
                'availableTrainingRowCount': 100,
                'validationDuration': 'P14D',
                'validationRowCount': 100,
                'gapStartDate': '2015-11-26T19:00:00.000000Z',
                'availableTrainingEndDate': '2015-11-26T19:00:00.000000Z',
                'validationStartDate': '2015-11-26T19:00:00.000000Z',
                'gapEndDate': '2015-11-26T19:00:00.000000Z',
                'gapDuration': 'P0Y0M0D',
                'gapRowCount': 100,
                'totalRowCount': 100
            }
        ],
        'datetimePartitionColumn': 'dates',
        'dateFormat': '%Y-%m-%d %H:%M:%S',
        'useTimeSeries': False,
        'defaultToAPriori': False,
        'defaultToKnownInAdvance': False,
        'featureDerivationWindowStart': None,
        'featureDerivationWindowEnd': None,
        'forecastWindowStart': None,
        'forecastWindowEnd': None,
        'projectId': project_id,
        'availableTrainingStartDate': '2015-06-08T20:00:00.000000Z',
        'primaryTrainingStartDate': '2015-11-10T19:00:00.000000Z',
        'holdoutEndDate': '2015-12-24T19:00:00.000000Z',
        'validationDuration': 'P14D',
        'availableTrainingDuration': 'P0Y0M184DT23H0M0S',
        'availableTrainingRowCount': 100,
        'numberOfBacktests': 1,
        'gapStartDate': '2015-12-10T19:00:00.000000Z',
        'availableTrainingEndDate': '2015-12-10T19:00:00.000000Z',
        'holdoutStartDate': holdout_start_date.isoformat(),
        'gapEndDate': '2015-12-10T19:00:00.000000Z',
        'holdoutDuration': 'P14D',
        'holdoutRowCount': 100,
        'gapDuration': 'P0Y0M0D',
        'gapRowCount': 100,
        'totalRowCount': 100,
        'autopilotDataSelectionMethod': 'rowCount'
    }


@pytest.fixture
def datetime_partition_after_target_server_data_feature_settings(
    datetime_partition_after_target_server_data
):
    settings = dict(datetime_partition_after_target_server_data)
    settings.update({
        'featureSettings': [
            {'featureName': 'timestamp', 'knownInAdvance': True, 'aPriori': True},
            {'featureName': 'target', 'knownInAdvance': True, 'aPriori': True},
            {'featureName': 'series_id', 'knownInAdvance': True, 'aPriori': True},
            {'featureName': 'input3', 'knownInAdvance': True, 'aPriori': True},
            {'featureName': 'input4', 'knownInAdvance': True, 'aPriori': True}
        ],
    })
    return settings


@pytest.fixture
def datetime_partition(datetime_partition_server_data):
    return DatetimePartitioning.from_server_data(datetime_partition_server_data)


@pytest.fixture
def datetime_partition_spec(datetime_partition_server_data, holdout_start_date):
    backtest_data = datetime_partition_server_data['backtests'][0]
    return DatetimePartitioningSpecification(
        datetime_partition_server_data['datetimePartitionColumn'],
        autopilot_data_selection_method=enums.DATETIME_AUTOPILOT_DATA_SELECTION_METHOD.ROW_COUNT,
        validation_duration=datetime_partition_server_data['validationDuration'],
        holdout_start_date=holdout_start_date,
        holdout_duration=datetime_partition_server_data['holdoutDuration'],
        gap_duration=datetime_partition_server_data['gapDuration'],
        number_of_backtests=datetime_partition_server_data['numberOfBacktests'],
        backtests=[BacktestSpecification(backtest_data['index'], backtest_data['gapDuration'],
                                         parse_time(backtest_data['validationStartDate']),
                                         backtest_data['validationDuration']
                                         )]
    )


@pytest.fixture
def datetime_partition_spec_feature_settings(
        datetime_partition_server_data_feature_settings, holdout_start_date
):
    backtest_data = datetime_partition_server_data_feature_settings['backtests'][0]
    feature_settings_data = datetime_partition_server_data_feature_settings['featureSettings']
    return DatetimePartitioningSpecification(
        datetime_partition_server_data_feature_settings['datetimePartitionColumn'],
        autopilot_data_selection_method=enums.DATETIME_AUTOPILOT_DATA_SELECTION_METHOD.ROW_COUNT,
        validation_duration=datetime_partition_server_data_feature_settings['validationDuration'],
        holdout_start_date=holdout_start_date,
        holdout_duration=datetime_partition_server_data_feature_settings['holdoutDuration'],
        gap_duration=datetime_partition_server_data_feature_settings['gapDuration'],
        number_of_backtests=datetime_partition_server_data_feature_settings['numberOfBacktests'],
        backtests=[BacktestSpecification(backtest_data['index'], backtest_data['gapDuration'],
                                         parse_time(backtest_data['validationStartDate']),
                                         backtest_data['validationDuration']
                                         )],
        feature_settings=[FeatureSettings(fs['featureName'], fs['knownInAdvance'])
                          for fs in feature_settings_data]
    )


@pytest.fixture
def datetime_partition_spec_bad_feature_settings(
        datetime_partition_server_data_feature_settings, holdout_start_date
):
    backtest_data = datetime_partition_server_data_feature_settings['backtests'][0]
    feature_settings_data = datetime_partition_server_data_feature_settings['featureSettings']
    return DatetimePartitioningSpecification(
        datetime_partition_server_data_feature_settings['datetimePartitionColumn'],
        autopilot_data_selection_method=enums.DATETIME_AUTOPILOT_DATA_SELECTION_METHOD.ROW_COUNT,
        validation_duration=datetime_partition_server_data_feature_settings['validationDuration'],
        holdout_start_date=holdout_start_date,
        holdout_duration=datetime_partition_server_data_feature_settings['holdoutDuration'],
        gap_duration=datetime_partition_server_data_feature_settings['gapDuration'],
        number_of_backtests=datetime_partition_server_data_feature_settings['numberOfBacktests'],
        backtests=[BacktestSpecification(backtest_data['index'], backtest_data['gapDuration'],
                                         parse_time(backtest_data['validationStartDate']),
                                         backtest_data['validationDuration']
                                         )],
        feature_settings=[FeatureSettings(fs['featureName'], None)
                          for fs in feature_settings_data]
    )


@pytest.fixture
def datetime_partition_spec_extra_time_series_controls(
        datetime_partition_server_data_extra_time_series_controls, holdout_start_date
):
    server_data = datetime_partition_server_data_extra_time_series_controls
    backtest_data = server_data['backtests'][0]
    return DatetimePartitioningSpecification(
        server_data['datetimePartitionColumn'],
        autopilot_data_selection_method=enums.DATETIME_AUTOPILOT_DATA_SELECTION_METHOD.ROW_COUNT,
        validation_duration=server_data['validationDuration'],
        holdout_start_date=holdout_start_date,
        holdout_duration=server_data['holdoutDuration'],
        gap_duration=server_data['gapDuration'],
        number_of_backtests=server_data['numberOfBacktests'],
        backtests=[BacktestSpecification(backtest_data['index'], backtest_data['gapDuration'],
                                         parse_time(backtest_data['validationStartDate']),
                                         backtest_data['validationDuration']
                                         )],
        treat_as_exponential=server_data['treatAsExponential'],
        differencing_method=server_data['differencingMethod'],
        periodicities=[Periodicity(p['timeSteps'], p['timeUnit'])
                       for p in server_data['periodicities']]
    )


@pytest.fixture
def time_series_partition_spec(datetime_partition_time_series_server_data, holdout_start_date):
    backtest_data = datetime_partition_time_series_server_data['backtests'][0]
    fdw_start = datetime_partition_time_series_server_data['featureDerivationWindowStart']
    fdw_end = datetime_partition_time_series_server_data['featureDerivationWindowEnd']
    return DatetimePartitioningSpecification(
        datetime_partition_time_series_server_data['datetimePartitionColumn'],
        autopilot_data_selection_method=enums.DATETIME_AUTOPILOT_DATA_SELECTION_METHOD.ROW_COUNT,
        validation_duration=datetime_partition_time_series_server_data['validationDuration'],
        holdout_start_date=holdout_start_date,
        holdout_duration=datetime_partition_time_series_server_data['holdoutDuration'],
        gap_duration=datetime_partition_time_series_server_data['gapDuration'],
        number_of_backtests=datetime_partition_time_series_server_data['numberOfBacktests'],
        backtests=[BacktestSpecification(backtest_data['index'], backtest_data['gapDuration'],
                                         parse_time(backtest_data['validationStartDate']),
                                         backtest_data['validationDuration']
                                         )],
        use_time_series=datetime_partition_time_series_server_data['useTimeSeries'],
        default_to_known_in_advance=datetime_partition_time_series_server_data[
            'defaultToKnownInAdvance'],
        feature_derivation_window_start=fdw_start,
        feature_derivation_window_end=fdw_end,
        forecast_window_start=datetime_partition_time_series_server_data['forecastWindowStart'],
        forecast_window_end=datetime_partition_time_series_server_data['forecastWindowEnd']
    )


@pytest.fixture
def datetime_project_server_data(project_id):
    return {
        'id': project_id,
        'projectName': 'Untitled Project',
        'fileName': 'special_date_kickcars.csv',
        'stage': 'modeling',
        'autopilotMode': 0,
        'created': '2017-01-05T17:13:51.257353Z',
        'target': 'y',
        'metric': 'LogLoss',
        'partition': {
            'datetimeCol': None,
            'cvMethod': 'datetime',
            'datetimePartitionColumn': 'Purch\xc3\xa4Date',
            'validationPct': None,
            'reps': None,
            'cvHoldoutLevel': None,
            'holdoutLevel': None,
            'userPartitionCol': None,
            'validationType': 'TVH',
            'trainingLevel': None,
            'partitionKeyCols': None,
            'holdoutPct': None,
            'validationLevel': None
        },
        'recommender': {
            'recommenderItemId': None,
            'isRecommender': None,
            'recommenderUserId': None
        },
        'advancedOptions': {
            'responseCap': False,
            'downsampledMinorityRows': None,
            'downsampledMajorityRows': None,
            'blueprintThreshold': 3,
            'seed': None,
            'weights': None,
            'smartDownsampled': False,
            'majorityDownsamplingRate': None
        },
        'positiveClass': 1,
        'maxTrainPct': 78.125,
        'holdoutUnlocked': False,
        'targetType': 'Binary'
    }


@pytest.fixture
def datetime_project(datetime_project_server_data):
    return Project.from_server_data(datetime_project_server_data)


@pytest.fixture
def datetime_model_server_data(project_id):
    return {
        'featurelistId': '586ef743ccf94e7d7310c288',
        'processes': [
            'Ordinal encoding of categorical variables',
            'Missing Values Imputed',
            'Gradient Boosted Trees Classifier'
        ],
        'featurelistName': 'Informative Features',
        'backtests': [
            {
                'index': 0,
                'score': 0.7,
                'status': 'COMPLETED',
                'training_start_date': '2015-10-26T19:00:00.000000Z',
                'training_duration': 'P0Y1M0D',
                'training_row_count': 880,
                'training_end_date': '2015-11-26T19:00:00.000000Z'
            },
            {
                'index': 1,
                'score': 0.4,
                'status': 'COMPLETED',
                'training_start_date': '2015-10-26T19:00:00.000000Z',
                'training_duration': 'P0Y1M0D',
                'training_row_count': 880,
                'training_end_date': '2015-11-26T19:00:00.000000Z'
            },
            {
                'index': 1,
                'score': None,
                'status': 'BOUNDARIES_EXCEEDED',
                'training_start_date': None,
                'training_duration': None,
                'training_row_count': None,
                'training_end_date': None
            }
        ],
        'modelType': 'Gradient Boosted Trees Classifier',
        'modelCategory': 'model',
        'projectId': project_id,
        'dataSelectionMethod': 'rowCount',
        'samplePct': None,
        'holdoutStatus': 'COMPLETED',
        'trainingStartDate': None,
        'metrics': {
            'AUC': {
                'backtesting': None,
                'holdout': None,
                'backtestingScores': [0.3, 0.23, None],
                'crossValidation': None,
                'validation': 0.75862
            },
            'Rate@Top5%': {
                'backtesting': None,
                'holdout': None,
                'backtestingScores': [1.0, 2.5, None],
                'crossValidation': None,
                'validation': 0.75
            },
            'Rate@TopTenth%': {
                'backtesting': None,
                'holdout': None,
                'backtestingScores': [1.223, 1.23, None],
                'crossValidation': None,
                'validation': 1
            },
            'RMSE': {
                'backtesting': None,
                'holdout': None,
                'backtestingScores': [1.5, 0.2, None],
                'crossValidation': None,
                'validation': 0.3721
            },
            'LogLoss': {
                'backtesting': None,
                'holdout': None,
                'backtestingScores': [0.7, 0.4, None],
                'crossValidation': None,
                'validation': 0.46308
            },
            'FVE Binomial': {
                'backtesting': None,
                'holdout': None,
                'backtestingScores': [0.9, 0.8, None],
                'crossValidation': None,
                'validation': 0.1381
            },
            'Gini Norm': {
                'backtesting': None,
                'holdout': None,
                'backtestingScores': [0.4, 0.3, None],
                'crossValidation': None,
                'validation': 0.51724
            },
            'Rate@Top10%': {
                'backtesting': None,
                'holdout': None,
                'backtestingScores': [0.5, 0.4, None],
                'crossValidation': None,
                'validation': 0.84615
            }
        },
        'trainingDuration': None,
        'trainingRowCount': 880,
        'holdoutScore': None,
        'trainingInfo': {
            'predictionTrainingDuration': 'P0Y1M0D',
            'holdoutTrainingStartDate': '2015-10-26T19:00:00.000000Z',
            'predictionTrainingStartDate': '2015-10-26T19:00:00.000000Z',
            'holdoutTrainingDuration': 'P0Y1M0D',
            'predictionTrainingRowCount': 880,
            'holdoutTrainingEndDate': '2015-12-10T19:00:00.000000Z',
            'predictionTrainingEndDate': '2015-12-10T19:00:00.000000Z',
            'holdoutTrainingRowCount': 880
        },
        'isFrozen': False,
        'blueprintId': '588d02115164b77809d1223820c8933f',
        'timeWindowSamplePct': None,
        'trainingEndDate': None,
        'id': '586ef75cccf94e7da010c294'
    }


@pytest.fixture
def datetime_start_end_model_server_data(datetime_model_server_data):
    final = dict(datetime_model_server_data)
    final['dataSelectionMethod'] = 'selectedDateRange'
    final['trainingRowCount'] = None
    final['trainingStartDate'] = '2015-10-26T19:00:00.000000Z'
    final['trainingEndDate'] = '2015-12-10T19:00:00.000000Z'
    final['timeWindowSamplePct'] = 50
    return final


@pytest.fixture
def datetime_model(datetime_model_server_data):
    return DatetimeModel.from_server_data(datetime_model_server_data)


@pytest.fixture
def datetime_model_job_server_data(project_id):
    body = {
        'status': 'inprogress',
        'processes': [
            'One-Hot Encoding',
            'Bernoulli Naive Bayes classifier (scikit-learn)',
            'Missing Values Imputed',
            'Gaussian Naive Bayes classifier (scikit-learn)',
            'Naive Bayes combiner classifier',
            'Calibrate predictions'
        ],
        'projectId': project_id,
        'samplePct': None,
        'modelType': 'Gradient Boosted Trees Classifier',
        'featurelistId': '586ef743ccf94e7d7310c288',
        'modelCategory': 'model',
        'blueprintId': '588d02115164b77809d1223820c8933f',
        'isBlocked': False,
        'id': '1'
    }
    return body



import json

import pytest
import responses
import trafaret

from datarobot import errors
from datarobot.models.training_predictions import RowsIterator


@responses.activate
def test_rows_iterator__empty_response__ok(client):
    body = {'data': [], 'next': None}
    responses.add(
        responses.GET,
        'https://host_name.com/foobar',
        body=json.dumps(body),
        content_type='application/json',
    )

    it = RowsIterator(client, 'foobar')

    assert list(it) == []


@responses.activate
def test_rows_iterator__two_items__ok(client):
    url = 'https://host_name.com/foobar'
    body = {'data': [{'foo': 'bar'}], 'next': '{}?offset=1&limit=1'.format(url)}
    responses.add(
        responses.GET,
        url,
        body=json.dumps(body),
        content_type='application/json',
    )

    it = RowsIterator(client, 'foobar', limit=1)

    assert next(it, None) == {'foo': 'bar'}
    assert len(responses.calls) == 1

    responses.reset()
    body2 = {'data': [{'bar': 'foo'}], 'next': '{}?offset=2&limit=1'.format(url)}
    responses.add(
        responses.GET,
        url,
        body=json.dumps(body2),
        content_type='application/json',
    )

    assert next(it, None) == {'bar': 'foo'}
    assert len(responses.calls) == 1

    responses.reset()
    body3 = {'data': [], 'next': '{}?offset=3&limit=1'.format(url)}
    responses.add(
        responses.GET,
        url,
        body=json.dumps(body3),
        content_type='application/json',
    )

    assert next(it, None) is None
    assert len(responses.calls) == 1


@responses.activate
def test_rows_iterator__http_504__raise_error(client):
    url = 'https://host_name.com/foobar'
    body = {'message': 'test-error'}
    responses.add(
        responses.GET,
        url,
        body=json.dumps(body),
        content_type='application/json',
        status=504,
    )

    with pytest.raises(errors.ServerError) as exc:
        it = RowsIterator(client, 'foobar')
        next(it, None)

    assert exc.value.status_code == 504


@responses.activate
def test_rows_iterator__no_data_field__raise_error(client):
    url = 'https://host_name.com/foobar'
    body = {'message': 'test-error', 'next': 'test'}
    responses.add(
        responses.GET,
        url,
        body=json.dumps(body),
        content_type='application/json',
    )

    with pytest.raises(Exception) as exc:
        it = RowsIterator(client, 'foobar')
        next(it, None)

    assert isinstance(exc.value, trafaret.DataError)



import logging
import sys


def get_logger(name):
    logger = logging.getLogger(name)
    logger.addHandler(logging.StreamHandler(sys.stdout))
    logger.setLevel(logging.INFO)
    logger.propagate = False
    return logger



from datarobot.utils import encode_utf8_if_py2


class EligibilityResult(object):
    """ Represents whether a particular operation is supported

    For instance, a function to check whether a set of models can be blended can return an
    EligibilityResult specifying whether or not blending is supported and why it may not be
    supported.

    Attributes
    ----------
    supported : bool
        whether the operation this result represents is supported
    reason : str
        why the operation is or is not supported
    context : str
        what operation isn't supported
    """
    def __init__(self, supported, reason='', context=''):
        self.supported = supported
        self.reason = reason
        self.context = context

    def __repr__(self):
        return encode_utf8_if_py2(u'{}(supported={}, reason=\'{}\', context=\'{}\')'.format(
            self.__class__.__name__, self.supported, self.reason, self.context)
        )



import mock
import responses
import six

from datarobot import DataDriver


@responses.activate
def test_list(drivers_endpoint, drivers_list_server_resp):
    responses.add(responses.GET, drivers_endpoint, json=drivers_list_server_resp)
    drivers = DataDriver.list()
    for driver, server_payload in zip(drivers, drivers_list_server_resp['data']):
        assert driver.id == server_payload['id']
        assert driver.class_name == server_payload['className']
        assert driver.base_names == server_payload['baseNames']
        assert driver.canonical_name == server_payload['canonicalName']
        assert driver.creator == server_payload['creator']


@responses.activate
def test_get(drivers_endpoint, postgresql_driver_server_resp):
    driver_id = postgresql_driver_server_resp['id']
    responses.add(
        responses.GET,
        '{}{}/'.format(drivers_endpoint, driver_id),
        json=postgresql_driver_server_resp
    )

    driver = DataDriver.get(driver_id)
    assert driver.id == postgresql_driver_server_resp['id']
    assert driver.class_name == postgresql_driver_server_resp['className']
    assert driver.base_names == postgresql_driver_server_resp['baseNames']
    assert driver.canonical_name == postgresql_driver_server_resp['canonicalName']
    assert driver.creator == postgresql_driver_server_resp['creator']


@responses.activate
def test_create(drivers_endpoint, driver_libraries_endpoint,
                driver_library_upload_server_resp, postgresql_driver_server_resp):
    responses.add(responses.POST, driver_libraries_endpoint,
                  json=driver_library_upload_server_resp)
    responses.add(responses.POST, drivers_endpoint, json=postgresql_driver_server_resp)

    with mock.patch('os.path.exists'):
        with mock.patch('datarobot.rest.open', create=True) as open_mock:
            open_mock.return_value = six.StringIO('thra\ntata\nrata')
            driver = DataDriver.create(
                class_name=postgresql_driver_server_resp['className'],
                canonical_name=postgresql_driver_server_resp['canonicalName'],
                files=['/tmp/{}'.format(name)
                       for name in postgresql_driver_server_resp['baseNames']]
            )
    assert driver.id == postgresql_driver_server_resp['id']
    assert driver.class_name == postgresql_driver_server_resp['className']
    assert driver.base_names == postgresql_driver_server_resp['baseNames']
    assert driver.canonical_name == postgresql_driver_server_resp['canonicalName']
    assert driver.creator == postgresql_driver_server_resp['creator']


@responses.activate
def test_create_local_multiple_libraries(
        drivers_endpoint, driver_libraries_endpoint,
        driver_library_upload_server_resp, postgresql_driver_server_resp
):
    postgresql_driver_server_resp['baseNames'] += postgresql_driver_server_resp['baseNames']
    responses.add(responses.POST, driver_libraries_endpoint,
                  json=driver_library_upload_server_resp)
    responses.add(responses.POST, drivers_endpoint, json=postgresql_driver_server_resp)

    with mock.patch('os.path.exists'):
        with mock.patch('datarobot.rest.open', create=True) as open_mock:
            open_mock.return_value = six.StringIO('thra\ntata\nrata')
            # send with multiple libraries
            driver = DataDriver.create(
                class_name=postgresql_driver_server_resp['className'],
                canonical_name=postgresql_driver_server_resp['canonicalName'],
                files=['/tmp/{}'.format(name)
                       for name in postgresql_driver_server_resp['baseNames']]
            )
    assert driver.id == postgresql_driver_server_resp['id']
    assert driver.class_name == postgresql_driver_server_resp['className']
    assert driver.base_names == postgresql_driver_server_resp['baseNames']
    assert driver.canonical_name == postgresql_driver_server_resp['canonicalName']
    assert driver.creator == postgresql_driver_server_resp['creator']


@responses.activate
def test_update(drivers_endpoint, postgresql_driver_server_resp):
    driver_id = postgresql_driver_server_resp['id']
    responses.add(
        responses.GET,
        '{}{}/'.format(drivers_endpoint, driver_id),
        json=postgresql_driver_server_resp
    )

    updated_postgresql_driver_server_resp = dict(
        postgresql_driver_server_resp,
        className='NewClassName',
        canonicalName='NewCanonicalName'
    )

    responses.add(
        responses.PATCH,
        '{}{}/'.format(drivers_endpoint, driver_id),
        json=updated_postgresql_driver_server_resp
    )

    driver = DataDriver.get(driver_id)
    driver.update(class_name='NewClassName', canonical_name='NewCanonicalName')
    assert driver.class_name == updated_postgresql_driver_server_resp['className']
    assert driver.canonical_name == updated_postgresql_driver_server_resp['canonicalName']


@responses.activate
def test_update_unable_to_change_base_names(
        drivers_endpoint, postgresql_driver_server_resp, mysql_driver_server_resp
):
    driver_id = postgresql_driver_server_resp['id']
    responses.add(
        responses.GET,
        '{}{}/'.format(drivers_endpoint, driver_id),
        json=postgresql_driver_server_resp
    )

    responses.add(
        responses.PATCH,
        '{}{}/'.format(drivers_endpoint, driver_id),
        json=postgresql_driver_server_resp
    )

    driver = DataDriver.get(driver_id)

    driver._base_names = mysql_driver_server_resp['baseNames']
    driver.update()

    # get driver from backend
    new_driver = DataDriver.get(driver_id)

    # assert server side has not been changed
    assert new_driver.base_names != mysql_driver_server_resp['baseNames']
    assert new_driver.base_names == postgresql_driver_server_resp['baseNames']


@responses.activate
def test_update_unable_to_change_creator(
        drivers_endpoint, postgresql_driver_server_resp, mysql_driver_server_resp
):
    driver_id = postgresql_driver_server_resp['id']
    responses.add(
        responses.GET,
        '{}{}/'.format(drivers_endpoint, driver_id),
        json=postgresql_driver_server_resp
    )

    responses.add(
        responses.PATCH,
        '{}{}/'.format(drivers_endpoint, driver_id),
        json=postgresql_driver_server_resp
    )

    driver = DataDriver.get(driver_id)

    driver._creator = mysql_driver_server_resp['creator']
    driver.update()

    # get driver from backend
    new_driver = DataDriver.get(driver_id)

    # assert server side has not been changed
    assert new_driver.creator != mysql_driver_server_resp['creator']
    assert new_driver.creator == postgresql_driver_server_resp['creator']


@responses.activate
def test_delete(drivers_endpoint, postgresql_driver_server_resp):
    driver_id = postgresql_driver_server_resp['id']
    responses.add(
        responses.GET,
        '{}{}/'.format(drivers_endpoint, driver_id),
        json=postgresql_driver_server_resp
    )
    delete_url = '{}{}/'.format(drivers_endpoint, driver_id)
    responses.add(responses.DELETE, delete_url)

    driver = DataDriver.get(driver_id)
    driver.delete()
    assert responses.calls[1].request.method == responses.DELETE
    assert responses.calls[1].request.url == delete_url



import json

import mock
import pytest
import responses
from datarobot import Job, ModelJob, Project, Blueprint, errors
from datarobot.models.modeljob import wait_for_async_model_creation
from tests.utils import SDKTestcase, assert_raised_regex


class TestWaitForAsyncModelCreation(SDKTestcase):
    mock_model_job_data = {
        'status': 'inprogress',
        'processes': [
            'One-Hot Encoding',
            'Bernoulli Naive Bayes classifier (scikit-learn)',
            'Missing Values Imputed',
            'Gaussian Naive Bayes classifier (scikit-learn)',
            'Naive Bayes combiner classifier',
            'Calibrate predictions'
        ],
        'projectId': 'p-id',
        'samplePct': 28.349,
        'modelType': 'Naive Bayes combiner classifier',
        'featurelistId': '56d8620bccf94e26f37af0a3',
        'modelCategory': 'model',
        'blueprintId': '2a1b9ae97fe61880332e196c770c1f9f',
        'isBlocked': False,
        'id': '1'
    }

    def setUp(self):
        super(TestWaitForAsyncModelCreation, self).setUp()
        self.pid = 'p-id'
        self.model_job_id = '5'
        self.get_job_url = 'https://host_name.com/projects/{}/modelJobs/{}/'.format(
            self.pid,
            self.model_job_id,
        )

    @responses.activate
    def test_timeout(self):
        responses.add(
            responses.GET,
            self.get_job_url,
            status=200,
            body=json.dumps(self.mock_model_job_data),
            content_type='application/json'
        )
        with pytest.raises(errors.AsyncTimeoutError) as exc_info:
            wait_for_async_model_creation(
                self.pid, self.model_job_id, max_wait=1)
        assert_raised_regex(exc_info, 'Model creation timed out in')

    @responses.activate
    @mock.patch('datarobot.models.modeljob.ModelJob.get_model')
    def test_success(self, get_model):
        responses.add(
            responses.GET,
            self.get_job_url,
            status=303,
            body='',
            content_type='application/json'
        )
        wait_for_async_model_creation(self.pid, self.model_job_id)
        self.assertEqual(get_model.call_count, 1)

    @responses.activate
    @mock.patch('datarobot.models.modeljob.ModelJob.get_model')
    def test_error(self, get_model):
        data = dict(self.mock_model_job_data, status='error')
        responses.add(
            responses.GET,
            self.get_job_url,
            status=200,
            body=json.dumps(data),
            content_type='application/json'
        )

        with pytest.raises(errors.AsyncModelCreationError) as exc_info:
            wait_for_async_model_creation(self.pid, self.model_job_id)
        assert_raised_regex(exc_info, "Model creation unsuccessful")


class TestJobModel(SDKTestcase):
    fully_trained_model = """
            {
        "featurelistId": "556cdfbd100d2b10048c7941",
        "processes": ["One", "Two", "Three"],
        "featurelistName": "Informative Features",
        "projectId": "p-id",
        "samplePct": 64,
        "modelType": "AVG Blender",
        "metrics": {
            "AUC": {
                "holdout": 0.76603,
                "validation": 0.64141,
                "crossValidation": 0.7625240000000001
            },
            "Rate@Top5%": {
                "holdout": 1,
                "validation": 0.5,
                "crossValidation": 0.9
            },
            "Rate@TopTenth%": {
                "holdout": 1,
                "validation": 1,
                "crossValidation": 1
            },
            "RMSE": {
                "holdout": 0.42054,
                "validation": 0.44396,
                "crossValidation": 0.40162000000000003
            },
            "LogLoss": {
                "holdout": 0.53707,
                "validation": 0.58051,
                "crossValidation": 0.5054160000000001
            },
            "FVE Binomial": {
                "holdout": 0.17154,
                "validation": 0.03641,
                "crossValidation": 0.17637399999999998
            },
            "Gini Norm": {
                "holdout": 0.53206,
                "validation": 0.28282,
                "crossValidation": 0.525048
            },
            "Rate@Top10%": {
                "holdout": 1,
                "validation": 0.25,
                "crossValidation": 0.7
            }
        },
        "blueprintId": "a4fd9d17a8ca62ee00590dd704dae6a8",
        "id": "5223deadbeefdeadbeef1234"
    }
    """

    final_model_location = ('https://host_name.com/projects/55666eb9100d2b109b59e267/'
                            'models/5223deadbeefdeadbeef1234/')

    mock_generic_job_data_inprogress = {
            'status': 'inprogress',
            'url': 'https://host_name.com/projects/p-id/modelJobs/1/',
            'id': '1',
            'jobType': 'model',
            'isBlocked': False,
            'projectId': 'p-id'
        }

    mock_generic_job_data_completed = dict(mock_generic_job_data_inprogress,
                                           status='COMPLETED')

    mock_model_job_data = {
        'status': 'inprogress',
        'processes': [
            'One-Hot Encoding',
            'Bernoulli Naive Bayes classifier (scikit-learn)',
            'Missing Values Imputed',
            'Gaussian Naive Bayes classifier (scikit-learn)',
            'Naive Bayes combiner classifier',
            'Calibrate predictions'
        ],
        'projectId': 'p-id',
        'samplePct': 28.349,
        'modelType': 'Naive Bayes combiner classifier',
        'featurelistId': '56d8620bccf94e26f37af0a3',
        'modelCategory': 'model',
        'blueprintId': '2a1b9ae97fe61880332e196c770c1f9f',
        'isBlocked': False,
        'id': '1'
    }

    mock_predict_job_data = {
        'status': 'error',
        'url': 'https://host_name.com/projects/p-id/predictJobs/64/',
        'id': '64',
        'jobType': 'predict',
        'isBlocked': False,
        'projectId': 'p-id'
    }

    def test_instantiate_job(self):
        data = {
            "status": "queue",
            "processes": [
                "One-Hot Encoding",
                "Missing Values Imputed",
                "RuleFit Classifier"
            ],
            "projectId": "556902e8100d2b3728d47551",
            "samplePct": 64,
            "modelType": "RuleFit Classifier",
            "featurelistId": "556902eb100d2b37d1130771",
            "blueprintId": "a8959bc1d46f07fb3dc14db7c1e3fc99",
            "isBlocked": False,
            "id": 11,
            "modelId": "556902ef100d2b37da13077d"
        }
        job = ModelJob(data)
        self.assertEqual(job.status, 'queue')
        self.assertEqual(job.job_type, 'model')
        self.assertEqual(job.processes, data['processes'])
        self.assertEqual(job.sample_pct, 64)
        self.assertEqual(job.model_type, "RuleFit Classifier")
        assert job.featurelist_id == data['featurelistId']
        self.assertIsInstance(job.project, Project)
        self.assertEqual(job.project.id, "556902e8100d2b3728d47551")
        self.assertIsInstance(job.blueprint, Blueprint)
        self.assertEqual(job.blueprint.id, "a8959bc1d46f07fb3dc14db7c1e3fc99")

        repr(job)

        with self.assertRaises(ValueError):
            ModelJob('qid')

    def test_future_proof(self):
        ModelJob(dict(self.mock_model_job_data, future='new'))

    @responses.activate
    def test_from_job(self):
        responses.add(
            responses.GET,
            'https://host_name.com/projects/p-id/modelJobs/1/',
            status=200,
            body=json.dumps(self.mock_model_job_data),
            content_type='application/json'
        )
        generic_job = Job(self.mock_generic_job_data_inprogress)
        model_job = ModelJob.from_job(generic_job)

        self.assertEqual(model_job.status, 'inprogress')
        self.assertEqual(model_job.job_type, 'model')
        self.assertEqual(model_job.project, Project('p-id'))
        self.assertEqual(model_job.processes, self.mock_model_job_data['processes'])

    def test_from_job_wrong_type(self):
        generic_job = Job(self.mock_predict_job_data)
        with self.assertRaises(ValueError):
            ModelJob.from_job(generic_job)

    @responses.activate
    def test_cancel(self):
        data = {'status': 'inprogress',
                'samplepct': 64.0,
                'processes': ['One-Hot Encoding',
                              'Missing Values Imputed',
                              'Standardize',
                              'Linear Regression'],
                'modelType': 'Linear Regression',
                'featurelistId': '55666d05100d2b01a1104dae',
                'blueprintId': '3bb4665320be633b30a9601b3e73284d',
                'projectId': '55666eb9100d2b109b59e267',
                'isBlocked': False,
                'id': 5,
                'modelId': '55666d0d100d2b01b1104db4'}
        job = ModelJob(data)
        responses.add(responses.DELETE,
                      'https://host_name.com/projects/55666eb9100d2b109b59e267/modelJobs/5/',
                      status=204,
                      body='',
                      content_type='application/json'
                      )
        job.cancel()
        self.assertEqual(responses.calls[0].request.method, 'DELETE')

    @responses.activate
    def test_get_success(self):
        data = {'status': 'inprogress',
                'samplepct': 64.0,
                'processes': ['One-Hot Encoding',
                              'Missing Values Imputed',
                              'Standardize',
                              'Linear Regression'],
                'modelType': 'Linear Regression',
                'featurelistId': '55666d05100d2b01a1104dae',
                'blueprintId': '3bb4665320be633b30a9601b3e73284d',
                'projectId': '55666eb9100d2b109b59e267',
                'isBlocked': False,
                'id': 5,
                'modelId': '55666d0d100d2b01b1104db4'}
        responses.add(responses.GET,
                      'https://host_name.com/projects/55666eb9100d2b109b59e267/modelJobs/5/',
                      status=200,
                      body=json.dumps(data),
                      content_type='application/json'
                      )
        job = ModelJob.get('55666eb9100d2b109b59e267', 5)
        self.assertEqual(responses.calls[0].request.method, 'GET')
        self.assertEqual(job.status, 'inprogress')

    @responses.activate
    def test_get_redirect(self):
        responses.add(
            responses.GET,
            'https://host_name.com/projects/55666eb9100d2b109b59e267/modelJobs/5/',
            content_type='application/json',
            status=303,
            body='',
            adding_headers={'Location': 'http://pam/api/v2/projects/p-id/models/id/'}
        )
        with self.assertRaises(errors.PendingJobFinished):
            ModelJob.get('55666eb9100d2b109b59e267', 5)

    @responses.activate
    def test_get_exceptional(self):
        responses.add(
            responses.GET,
            'https://host_name.com/projects/55666eb9100d2b109b59e267/modelJobs/5/',
            status=404,
            body=''
        )
        with self.assertRaises(errors.AppPlatformError):
            ModelJob.get('55666eb9100d2b109b59e267', 5)

    @responses.activate
    def test_get_unexpected_status_code(self):
        responses.add(
            responses.GET,
            'https://host_name.com/projects/55666eb9100d2b109b59e267/modelJobs/5/',
            status=202,
            body=''
        )
        with pytest.raises(errors.AsyncFailureError) as exc_info:
            ModelJob.get('55666eb9100d2b109b59e267', 5)
        assert_raised_regex(exc_info, "Server unexpectedly returned status code")

    @responses.activate
    def test_get_model_success(self):
        responses.add(
            responses.GET,
            'https://host_name.com/projects/55666eb9100d2b109b59e267/modelJobs/5/',
            status=303,
            body='',
            adding_headers={'Location': self.final_model_location}
        )
        responses.add(
            responses.GET,
            self.final_model_location,
            status=200,
            body=self.fully_trained_model
        )
        ModelJob.get_model('55666eb9100d2b109b59e267', 5)

    @responses.activate
    def test_get_model_from_job_success(self):
        generic_job = Job(self.mock_generic_job_data_inprogress)
        responses.add(
            responses.GET,
            'https://host_name.com/projects/p-id/jobs/1/',
            status=303,
            body=json.dumps(self.mock_generic_job_data_completed),
            adding_headers={'Location': self.final_model_location}
        )
        responses.add(
            responses.GET,
            self.final_model_location,
            status=200,
            body=self.fully_trained_model
        )
        generic_job.get_result()

    @responses.activate
    def test_get_model_from_job_not_finished(self):
        generic_job = Job(self.mock_generic_job_data_inprogress)
        responses.add(
            responses.GET,
            'https://host_name.com/projects/p-id/jobs/1/',
            status=200,
            body=json.dumps(self.mock_generic_job_data_inprogress),
            adding_headers={'Location': self.final_model_location}
        )
        with self.assertRaises(errors.JobNotFinished):
            generic_job.get_result()

    @responses.activate
    def test_get_model_not_finished(self):
        model_job_data = {
            'status': 'queue',
            'processes': [
                'One-Hot Encoding',
                'Missing Values Imputed',
                'RuleFit Classifier'
            ],
            'projectId': '556902e8100d2b3728d47551',
            'samplePct': 64,
            'modelType': 'RuleFit Classifier',
            'featurelistId': '556902eb100d2b37d1130771',
            'blueprintId': 'a8959bc1d46f07fb3dc14db7c1e3fc99',
            'isBlocked': False,
            'id': 5,
            'modelId': '556902ef100d2b37da13077d'
        }

        responses.add(
            responses.GET,
            'https://host_name.com/projects/55666eb9100d2b109b59e267/modelJobs/5/',
            status=200,
            body=json.dumps(model_job_data)
        )
        with self.assertRaises(errors.JobNotFinished):
            ModelJob.get_model('55666eb9100d2b109b59e267', 5)

    @responses.activate
    def test_get_model_exceptional(self):
        responses.add(
            responses.GET,
            'https://host_name.com/projects/55666eb9100d2b109b59e267/modelJobs/5/',
            status=404,
            body=''
        )
        with self.assertRaises(errors.AppPlatformError):
            ModelJob.get_model('55666eb9100d2b109b59e267', 5)

    @responses.activate
    def test_get_model_unexpected_status_code(self):
        final_model_location = ('https://host_name.com/projects/55666eb9100d2b109b59e267/'
                                'models/5223deadbeefdeadbeef1234/')

        responses.add(
            responses.GET,
            'https://host_name.com/projects/55666eb9100d2b109b59e267/modelJobs/5/',
            status=202,
            body='',
            adding_headers={'Location': final_model_location}
        )

        with pytest.raises(errors.AsyncFailureError) as exc_info:
            ModelJob.get_model('55666eb9100d2b109b59e267', 5)
        assert_raised_regex(exc_info, 'Server unexpectedly returned status code')



import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.utils import encode_utf8_if_py2


class ConfusionChart(APIObject):
    """ Confusion Chart data for model.



    Attributes
    ----------
    source : str
        Confusion Chart data source. Can be 'validation', 'crossValidation' or 'holdout'.
    raw_data : dict
        All of the raw data for the Confusion Chart
    confusion_matrix : list of list
        The NxN confusion matrix
    classes : list
        The names of each of the classes
    class_metrics : list of dict
        Containing all of the metrics for each of the classes. Dictionary keys:
            className : string
                name of the class
            actualCount : int
                number of times this class is seen in the validation data
            predictedCount : int
                number of times this class has been predicted for the validation data
            f1 : float
                F1 score
            recall : float
                recall score
            precision : float
                precision score
            wasActualPercentages : list of dict
                one vs all actual percentages in a format specified below, Dictionary keys:
                    otherClassName : string
                         the name of the other class
                    percentage : float
                         the percentage of the times this class was predicted when is was actually
                         class (from 0 to 1)
            wasPredictedPercentages : list of dict
                one vs all predicted percentages in a format specified below, Dictionary keys:
                    otherClassName : string
                        the name of the other class
                    percentage : float
                        the percentage of the times this class was actual predicted (from 0 to 1)
            confusionMatrixOneVsAll : list of list
                2d list representing 2x2 one vs all matrix. This represents the True/False
                Negative/Positive rates as integer for each class. The data structure looks like:
                ``[ [ True Negative, False Positive ], [ False Negative, True Positive ] ]``
    """

    _PercentageFloat = t.Float(gte=0, lte=1)  # 0.0 <= x <= 1.0

    _ClassPercentage = t.Dict({
        t.Key('other_class_name'): t.String,
        t.Key('percentage'): _PercentageFloat,
    }).ignore_extra('*')

    _ClassMetrics = t.Dict({
        t.Key('class_name'): t.String,
        # blue bars on the axis
        t.Key('actual_count'): t.Int,
        t.Key('predicted_count'): t.Int,
        # one vs all metrics
        t.Key('f1'): _PercentageFloat,
        t.Key('recall'): _PercentageFloat,
        t.Key('precision'): _PercentageFloat,
        t.Key('confusion_matrix_one_vs_all'): t.List(t.List(t.Int)),
        t.Key('was_actual_percentages'): t.List(_ClassPercentage),
        t.Key('was_predicted_percentages'): t.List(_ClassPercentage),
    }).ignore_extra('*')

    _ConfusionChartTrafaret = t.Dict({
        # available classes
        t.Key('classes'): t.List(t.String),
        # NxN confusion matrix
        t.Key('confusion_matrix'): t.List(t.List(t.Int)),
        t.Key('class_metrics'): t.List(_ClassMetrics),
    }).ignore_extra('*')

    ConfusionChartWrapper = t.Dict({
        t.Key('data'): _ConfusionChartTrafaret,
        t.Key('source'): t.String,
    }).ignore_extra('*')

    _converter = ConfusionChartWrapper

    def __init__(self, source, data):
        self.source = source
        self.raw_data = data
        self.class_metrics = data['class_metrics']
        self.confusion_matrix = data['confusion_matrix']
        self.classes = data['classes']

    def __repr__(self):
        return encode_utf8_if_py2(u'ConfusionChart({})'.format(self.source))



# flake8: noqa
# because the unused imports are on purpose

from .model import (Model, PrimeModel, BlenderModel, DatetimeModel, FrozenModel, RatingTableModel,
                    ModelParameters)
from .modeljob import ModelJob
from .blueprint import Blueprint, BlueprintChart, BlueprintTaskDocument, ModelBlueprintChart
from .predict_job import PredictJob
from .featurelist import Featurelist, ModelingFeaturelist
from .feature import Feature, ModelingFeature
from .job import Job, TrainingPredictionsJob
from .project import Project
from .prediction_dataset import PredictionDataset
from .prime_file import PrimeFile
from .ruleset import Ruleset
from .imported_model import ImportedModel
from .reason_codes import ReasonCodesInitialization, ReasonCodes
from .prediction_explanations import PredictionExplanationsInitialization, PredictionExplanations
from .rating_table import RatingTable
from .training_predictions import TrainingPredictions
from .recommended_model import ModelRecommendation
from .driver import DataDriver
from .data_store import DataStore
from .data_source import DataSource, DataSourceParameters
from .predictions import Predictions



import time

from datarobot import errors


def wait_for_async_resolution(client, async_location, max_wait=600):
    """
    Wait for successful resolution of the provided async_location.

    Parameters
    ----------
    client : RESTClientObject
        The configured v2 requests session
    async_location : str
        The URL we are polling for resolution. This can be either a fully-qualified URL
        like `http://host.com/routeName/` or just the relative route within the API
        i.e. `routeName/`.
    max_wait : int
        The number of seconds to wait before giving up

    Returns
    -------
    location : str
        The URL of the now-finished resource

    Raises
    ------
    AsyncFailureError
        If any of the responses from the server are unexpected
    AsyncProcessUnsuccessfulError
        If the job being waited for has failed or has been cancelled.
    AsyncTimeoutError
        If the resource did not resolve in time
    """
    start_time = time.time()

    join_endpoint = not async_location.startswith('http')  # Accept full qualified and relative urls

    response = client.get(async_location, allow_redirects=False, join_endpoint=join_endpoint)
    while time.time() < start_time + max_wait:
        if response.status_code == 303:
            return response.headers['Location']
        if response.status_code != 200:
            e_template = 'The server gave an unexpected response. Status Code {}: {}'
            raise errors.AsyncFailureError(e_template.format(
                response.status_code, response.text))
        data = response.json()
        if data['status'].lower()[:5] in ['error', 'abort']:
            e_template = 'The job did not complete successfully. Job Data: {}'
            raise errors.AsyncProcessUnsuccessfulError(e_template.format(data))
        time.sleep(5)
        response = client.get(async_location, allow_redirects=False, join_endpoint=join_endpoint)

    timeout_msg = 'Client timed out in {} seconds waiting for {} to resolve. Last status was {}: {}'
    raise errors.AsyncTimeoutError(timeout_msg.format(max_wait,
                                                      async_location,
                                                      response.status_code,
                                                      response.text))



import os

import pytest
import responses

from datarobot import Model
from datarobot.errors import ClientError


@pytest.fixture
def scoring_code_file_contents():
    # Just huge string bigger then 1MB chunk size
    return 'some_random_file_contents' * 1000000


@pytest.fixture
def local_jar_filename(tmpdir):
    return str(tmpdir.join('temp.jar'))


@responses.activate
@pytest.mark.usefixtures('client')
def test_positive_case(project_id, model_id, scoring_code_file_contents, local_jar_filename):
    url = 'https://host_name.com/projects/{}/models/{}/scoringCode/?sourceCode=false'.format(
        project_id, model_id)
    responses.add(responses.GET,
                  url,
                  body=scoring_code_file_contents,
                  status=200,
                  match_querystring=True)
    model = Model(project_id=project_id, id=model_id)
    model.download_scoring_code(local_jar_filename)
    assert os.path.isfile(local_jar_filename)
    with open(local_jar_filename) as f:
        contents = f.read()
        assert contents == scoring_code_file_contents
    os.unlink(local_jar_filename)
    assert not os.path.isfile(local_jar_filename)


@responses.activate
@pytest.mark.usefixtures('client')
def test_negative_case(project_id, model_id, local_jar_filename):
    url = 'https://host_name.com/projects/{}/models/{}/scoringCode/?sourceCode=false'.format(
        project_id, model_id)
    error_message = 'some error'
    responses.add(responses.GET,
                  url,
                  json={'message': error_message},
                  status=404,
                  match_querystring=True)
    model = Model(project_id=project_id, id=model_id)
    with pytest.raises(ClientError) as exc_info:
        model.download_scoring_code(local_jar_filename)
    assert not os.path.isfile(local_jar_filename)
    assert error_message in str(exc_info.value)



import json

import pytest
import responses


from datarobot import Model
from datarobot.errors import ClientError
from datarobot.models.pareto_front import ParetoFront, Solution


@pytest.fixture
def pareto_front_data(project_id):
    return {
        'project_id': project_id,
        'error_metric': 'Log Loss Error',
        'hyperparameters': {},
        'target_type': 'Binary',
        'solutions': [
            {
                'eureqa_solution_id': '5b2940f202d5abf17ddca9f8',
                'complexity': 1,
                'error': 0.6568885703145153,
                'expression': 'target = 0',
                'expression_annotated': 'target = 0',
                'best_model': False
            },
            {
                'eureqa_solution_id': '5b2940f202d5abf17ddca9f9',
                'complexity': 2,
                'error': 0.6463236186153043,
                'expression': 'target = 0',
                'expression_annotated': 'target = 0',
                'best_model': False
            },
            {
                'eureqa_solution_id': '5b2940f202d5abf17ddca9fa',
                'complexity': 3,
                'error': 0.6369812561414678,
                'expression': 'target = 0',
                'expression_annotated': 'target = 0',
                'best_model': False
            },
            {
                'eureqa_solution_id': '5b2940f202d5abf17ddca9fb',
                'complexity': 4,
                'error': 0.6341910672553971,
                'expression': 'target = 0',
                'expression_annotated': 'target = 0',
                'best_model': False
            },
            {
                'eureqa_solution_id': '5b29406c6523cd0665685a8d',
                'complexity': 5,
                'error': 0.6327990044770845,
                'expression': 'target = 0',
                'expression_annotated': 'target = 0',
                'best_model': True
            }
        ]
    }


@pytest.fixture
def server_pareto_front_data(project_id):
    return {
        'projectId': project_id,
        'errorMetric': 'Log Loss Error',
        'hyperparameters': {},
        'targetType': 'Binary',
        'solutions': [
            {
                'eureqaSolutionId': '5b2940f202d5abf17ddca9f8',
                'complexity': 1,
                'error': 0.6568885703145153,
                'expression': 'target = 0',
                'expressionAnnotated': 'target = 0',
                'bestModel': False
            },
            {
                'eureqaSolutionId': '5b2940f202d5abf17ddca9f9',
                'complexity': 2,
                'error': 0.6463236186153043,
                'expression': 'target = 0',
                'expressionAnnotated': 'target = 0',
                'bestModel': False
            },
            {
                'eureqaSolutionId': '5b2940f202d5abf17ddca9fa',
                'complexity': 3,
                'error': 0.6369812561414678,
                'expression': 'target = 0',
                'expressionAnnotated': 'target = 0',
                'bestModel': False
            },
            {
                'eureqaSolutionId': '5b2940f202d5abf17ddca9fb',
                'complexity': 4,
                'error': 0.6341910672553971,
                'expression': 'target = 0',
                'expressionAnnotated': 'target = 0',
                'bestModel': False
            },
            {
                'eureqaSolutionId': '5b29406c6523cd0665685a8d',
                'complexity': 5,
                'error': 0.6327990044770845,
                'expression': 'target = 0',
                'expressionAnnotated': 'target = 0',
                'bestModel': True
            }
        ]
    }


def test_instantiation(server_pareto_front_data, project_id):
    pareto_front = ParetoFront.from_server_data(server_pareto_front_data)

    assert pareto_front.project_id == project_id
    assert pareto_front.error_metric == 'Log Loss Error'
    assert pareto_front.hyperparameters == {}
    assert pareto_front.target_type == 'Binary'


def test_future_proof(server_pareto_front_data, project_id):
    data_with_future_keys = dict(server_pareto_front_data, new_key='amazing new key')
    data_with_future_keys['solutions'][0]['new_key'] = 'amazing new key'
    pareto_front = ParetoFront.from_server_data(data_with_future_keys)

    assert pareto_front.project_id == project_id
    assert pareto_front.error_metric == 'Log Loss Error'
    assert pareto_front.hyperparameters == {}
    assert pareto_front.target_type == 'Binary'


@pytest.fixture
def pareto_front_data_url(project_id, model_id):
    return 'https://host_name.com/projects/{}/eureqaModels/{}/'.format(
        project_id,
        model_id
    )


@responses.activate
@pytest.mark.usefixtures('client')
def test_get_pareto_front(server_pareto_front_data,
                          pareto_front_data,
                          pareto_front_data_url,
                          project_id, model_id):
    responses.add(
        responses.GET,
        pareto_front_data_url,
        status=200,
        content_type='application/json',
        body=json.dumps(server_pareto_front_data)
    )
    model = Model(id=model_id, project_id=project_id)
    pf = model.get_pareto_front()
    mock_data = pareto_front_data

    assert pf.project_id == project_id
    assert pf.error_metric == 'Log Loss Error'
    assert pf.hyperparameters == {}
    assert pf.target_type == 'Binary'

    for i in range(0, 5):
        assert pf.solutions[i].__dict__ == Solution(
            project_id=project_id, **mock_data['solutions'][i]).__dict__


@responses.activate
@pytest.mark.usefixtures('client')
def test_non_eureqa_model(pareto_front_data_url,
                          project_id, model_id):
    responses.add(
        responses.GET,
        pareto_front_data_url,
        status=404,
        content_type='application/json',
        body=json.dumps({})
    )
    model = Model(id=model_id, project_id=project_id)

    with pytest.raises(ClientError):
        model.get_pareto_front()



import six
import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.utils import deprecation_warning, from_api, to_api
from ..utils import encode_utf8_if_py2, parse_time


class BaseFeaturelist(APIObject):
    _base_path = None  # path, to be overridden in inheriting classes
    # nulls are to account for the partial featurelist created inside models
    _converter = t.Dict({
        t.Key('id'): t.Or(t.String, t.Null),
        t.Key('name'): t.Or(t.String, t.Null),
        t.Key('features'): t.Or(t.List(t.String), t.Null),
        t.Key('project_id'): t.Or(t.String(), t.Null),
        t.Key('created'): t.Or(t.Null, t.Call(parse_time)),
        t.Key('is_user_created'): t.Or(t.Bool, t.Null),
        t.Key('num_models'): t.Or(t.Int, t.Null),
        t.Key('description'): t.String(allow_blank=True)
    }).ignore_extra("*")

    def __init__(self, id=None, name=None, features=None, project_id=None, created=None,
                 is_user_created=None, num_models=None, description=None):
        self.id = id
        self.project_id = project_id
        self.name = name
        self.features = features
        self.created = created
        self.is_user_created = is_user_created
        self.num_models = num_models
        self.description = description

    def __repr__(self):
        return encode_utf8_if_py2(u'{}({})'.format(self.__class__.__name__, self.name))

    def _make_url(self):
        return '{}{}/'.format(self._base_path.format(self.project_id), self.id)

    def update(self, name=None, description=None):
        """ Update the name or description of an existing featurelist

        Note that only user-created featurelists can be renamed, and that names must not
        conflict with names used by other featurelists.

        Parameters
        ----------
        name : str, optional
            the new name for the featurelist
        description : str, optional
            the new description for the featurelist
        """
        data = {}
        if name is not None:
            data['name'] = name
        if description is not None:
            data['description'] = description
        self._client.patch(self._make_url(), data=data)
        if name is not None:
            self.name = name
        if description is not None:
            self.description = description

    def delete(self, dry_run=False, delete_dependencies=False):
        """ Delete a featurelist, and any models and jobs using it

        All models using a featurelist, whether as the training featurelist or as a monotonic
        constraint featurelist, will also be deleted when the deletion is executed and any queued or
        running jobs using it will be cancelled. Similarly, predictions made on these models will
        also be deleted. All the entities that are to be deleted with a featurelist are described
        as "dependencies" of it.  To preview the results of deleting a featurelist, call delete
        with `dry_run=True`

        When deleting a featurelist with dependencies, users must specify `delete_dependencies=True`
        to confirm they want to delete the featurelist and all its dependencies. Without that
        option, only featurelists with no dependencies may be successfully deleted and others will
        error.

        Featurelists configured into the project as a default featurelist or as a default monotonic
        constraint featurelist cannot be deleted.

        Featurelists used in a model deployment cannot be deleted until the model deployment is
        deleted.

        Parameters
        ----------
        dry_run : bool, optional
            specify True to preview the result of deleting the featurelist, instead of actually
            deleting it.
        delete_dependencies : bool, optional
            specify True to successfully delete featurelists with dependencies; if left
            False by default, featurelists without dependencies can be successfully deleted and
            those with dependencies will error upon attempting to delete them.

        Returns
        -------
        result : dict
            A dictionary describing the result of deleting the featurelist, with the following keys
                - dry_run : bool, whether the deletion was a dry run or an actual deletion
                - can_delete : bool, whether the featurelist can actually be deleted
                - deletion_blocked_reason : str, why the featurelist can't be deleted (if it can't)
                - num_affected_models : int, the number of models using this featurelist
                - num_affected_jobs : int, the number of jobs using this featurelist
        """
        result = self._client.delete(self._make_url(),
                                     params=to_api({'dry_run': dry_run,
                                                    'delete_dependencies': delete_dependencies}))
        checker = t.Dict({t.Key('dry_run'): t.Bool(), t.Key('can_delete'): t.Bool(),
                          t.Key('deletion_blocked_reason'): t.String(allow_blank=True),
                          t.Key('num_affected_models'): t.Int(),
                          t.Key('num_affected_jobs'): t.Int()}).ignore_extra('*')
        return checker.check(from_api(result.json()))


class Featurelist(BaseFeaturelist):
    """ A set of features used in modeling

    Attributes
    ----------
    id : str
        the id of the featurelist
    name : str
        the name of the featurelist
    features : list of str
        the names of all the Features in the featurelist
    project_id : str
        the project the featurelist belongs to
    created : datetime.datetime
        (New in version v2.13) when the featurelist was created
    is_user_created : bool
        (New in version v2.13) whether the featurelist was created by a user or by DataRobot
        automation
    num_models : int
        (New in version v2.13) the number of models currently using this featurelist.  A model is
        considered to use a featurelist if it is used to train the model or as a monotonic
        constraint featurelist, or if the model is a blender with at least one component model
        using the featurelist.
    description : basestring
        (New in version v2.13) the description of the featurelist.  Can be updated by the user
        and may be supplied by default for DataRobot-created featurelists.
    """
    _base_path = 'projects/{}/featurelists/'

    def __init__(self, id=None, name=None, features=None, project_id=None, created=None,
                 is_user_created=None, num_models=None, description=None):
        if isinstance(id, dict):
            deprecation_warning('Featurelist instantiation from a dict',
                                deprecated_since_version='v2.3',
                                will_remove_version='v3.0',
                                )
            super(Featurelist, self).__init__(**id)
        else:
            super(Featurelist, self).__init__(
                id=id, name=name, features=features, project_id=project_id, created=created,
                is_user_created=is_user_created, num_models=num_models, description=description
            )
        self._project = None
        self._set_up(self.project_id)

    @property
    def project(self):
        deprecation_warning('Featurelist.project',
                            deprecated_since_version='v2.3',
                            will_remove_version='v3.0',
                            message='Users should construct a Project from the project_id')
        return self._project

    def _set_up(self, project_id):
        from . import Project
        self._project = Project(project_id)

    @classmethod
    def from_data(cls, data):
        """ Overrides the parent method to ensure description is always populated

        Parameters
        ----------
        data : dict
            the data from the server, having gone through processing
        """
        # these keys are sometimes missing b/c dummy featurelists are instantiated inside models
        # in _normal_ usage these should all be supplied
        for key in ['features', 'num_models', 'is_user_created', 'created', 'id', 'name',
                    'project_id']:
            data.setdefault(key, None)
        if 'description' not in data:
            data['description'] = ''
        return super(Featurelist, cls).from_data(data)

    @classmethod
    def get(cls, project_id, featurelist_id):
        """Retrieve a known feature list

        Parameters
        ----------
        project_id : str
            The id of the project the featurelist is associated with
        featurelist_id : str
            The ID of the featurelist to retrieve

        Returns
        -------
        featurelist : Featurelist
            The queried instance
        """
        from . import Project

        if isinstance(project_id, Project):
            deprecation_warning('Featurelist.get using an instance of Project',
                                deprecated_since_version='v2.3',
                                will_remove_version='v3.0',
                                message='Use a project_id string instead')
            project_id = project_id.id
        elif isinstance(project_id, six.string_types):
            pass
        else:
            raise ValueError('Project arg must be Project instance or str')

        url = '{}{}/'.format(cls._base_path.format(project_id), featurelist_id)
        return cls.from_location(url)


class ModelingFeaturelist(BaseFeaturelist):
    """ A set of features that can be used to build a model

    In time series projects, a new set of modeling features is created after setting the
    partitioning options.  These features are automatically derived from those in the project's
    dataset and are the features used for modeling.  Modeling features are only accessible once
    the target and partitioning options have been set.  In projects that don't use time series
    modeling, once the target has been set, ModelingFeaturelists and Featurelists will behave
    the same.

    For more information about input and modeling features, see the
    :ref:`time series documentation<input_vs_modeling>`.

    Attributes
    ----------
    id : str
        the id of the modeling featurelist
    project_id : str
        the id of the project the modeling featurelist belongs to
    name : str
        the name of the modeling featurelist
    features : list of str
        a list of the names of features included in this modeling featurelist
    created : datetime.datetime
        (New in version v2.13) when the featurelist was created
    is_user_created : bool
        (New in version v2.13) whether the featurelist was created by a user or by DataRobot
        automation
    num_models : int
        (New in version v2.13) the number of models currently using this featurelist.  A model is
        considered to use a featurelist if it is used to train the model or as a monotonic
        constraint featurelist, or if the model is a blender with at least one component model
        using the featurelist.
    description : basestring
        (New in version v2.13) the description of the featurelist.  Can be updated by the user
        and may be supplied by default for DataRobot-created featurelists.
    """
    _base_path = 'projects/{}/modelingFeaturelists/'

    @classmethod
    def get(cls, project_id, featurelist_id):
        """ Retrieve a modeling featurelist

        Modeling featurelists can only be retrieved once the target and partitioning options have
        been set.

        Parameters
        ----------
        project_id : str
            the id of the project the modeling featurelist belongs to
        featurelist_id : str
             the id of the modeling featurelist to retrieve

        Returns
        -------
        featurelist : ModelingFeaturelist
            the specified featurelist
        """
        url = '{}{}/'.format(cls._base_path.format(project_id), featurelist_id)
        return cls.from_location(url)



import warnings


class AppPlatformError(Exception):
    """
    Raised by :meth:`Client.request()` for requests that:
      - Return a non-200 HTTP response, or
      - Connection refused/timeout or
      - Response timeout or
      - Malformed request
      - Have a malformed/missing header in the response.
    """
    def __init__(self, exc_message, status_code, error_code=None, json=None):
        super(AppPlatformError, self).__init__(exc_message)
        self.status_code = status_code
        self.error_code = error_code
        self.json = json or {}


class ServerError(AppPlatformError):
    """
    For 500-level responses from the server
    """


class ClientError(AppPlatformError):
    """
    For 400-level responses from the server
    has json parameter for additional information to be stored about error
    if need be
    """


class InputNotUnderstoodError(Exception):
    """
    Raised if a method is called in a way that cannot be understood
    """


class InvalidUsageError(Exception):
    """ Raised when methods are called with invalid or incompatible arguments
    """


class AllRetriesFailedError(Exception):
    """Raised when the retry manager does not successfully make a request"""


class InvalidModelCategoryError(Exception):
    """
    Raised when method specific for model category was called from wrong model
    """


class AsyncTimeoutError(Exception):
    """
    Raised when an asynchronous operation did not successfully get resolved
    within a specified time limit
    """


class AsyncFailureError(Exception):
    """
    Raised when querying an asynchronous status resulted in an exceptional
    status code (not 200 and not 303)
    """


class ProjectAsyncFailureError(AsyncFailureError):
    """
    When an AsyncFailureError occurs during project creation or finalizing the project
    settings for modeling. This exception will have the attributes ``status_code``
    indicating the unexpected status code from the server, and ``async_location`` indicating
    which asynchronous status object was being polled when the failure happened.
    """
    def __init__(self, exc_message, status_code, async_location):
        super(ProjectAsyncFailureError, self).__init__(exc_message)
        self.status_code = status_code
        self.async_location = async_location


class AsyncProcessUnsuccessfulError(Exception):
    """
    Raised when querying an asynchronous status showed that async process
    was not successful
    """


class AsyncModelCreationError(Exception):
    """
    Raised when querying an asynchronous status showed that model creation
    was not successful
    """


class AsyncPredictionsGenerationError(Exception):
    """
    Raised when querying an asynchronous status showed that predictions
    generation was not successful
    """


class PendingJobFinished(Exception):
    """
    Raised when the server responds with a 303 for the pending creation of a
    resource.
    """


class JobNotFinished(Exception):
    """
    Raised when execution was trying to get a finished resource from a pending
    job, but the job is not finished
    """


class DuplicateFeaturesError(Exception):
    """
    Raised when trying to create featurelist with duplicating features
    """


class DataRobotDeprecationWarning(DeprecationWarning):
    """
    Raised when using deprecated functions or using functions in a deprecated way
    """
    pass


class IllegalFileName(Exception):
    """
    Raised when trying to use a filename we can't handle.
    """


class JobAlreadyRequested(ClientError):
    """
    Raised when the requested model has already been requested.
    """
    pass


class InvalidRatingTableWarning(Warning):
    """
    Raised when using interacting with rating tables that failed validation
    """
    pass


class NoRedundancyImpactAvailable(Warning):
    """ Raised when retrieving old feature impact data

    Redundancy detection was added in v2.13 of the API, and some projects, e.g. multiclass projects
    do not support redundancy detection. This warning is raised to make
    clear that redundancy detection is unavailable.
    """


warnings.filterwarnings('default', category=DataRobotDeprecationWarning)
warnings.filterwarnings('always', category=InvalidRatingTableWarning)



import json

import pytest
import responses
from six.moves.urllib_parse import urlparse, parse_qs

from datarobot import Feature, ModelingFeature, Project

from trafaret import DataError


@pytest.fixture
def list_features_json():
    return """
    [{
        "id": 34,
        "name": "Claim_Amount",
        "featureType": "Numeric",
        "importance": 1,
        "lowInformation": false,
        "uniqueCount": 200,
        "naCount": 0,
        "dateFormat": null,
        "projectId": "556cdfbb100d2b0e88585195",
        "min": 1,
        "max": 1,
        "mean": 1,
        "median": 1,
        "stdDev": 0,
        "timeSeriesEligible": false,
        "timeSeriesEligibilityReason": "notADate",
        "timeStep": null,
        "timeUnit": null,
        "targetLeakage": "SKIPPED_DETECTION"
    }]
    """


@pytest.fixture
def feature_json():
    return """
    {
        "id": 34,
        "name": "Claim_Amount",
        "featureType": "Numeric",
        "importance": 1,
        "lowInformation": false,
        "uniqueCount": 200,
        "naCount": 0,
        "dateFormat": null,
        "projectId": "556cdfbb100d2b0e88585195",
        "min": 1,
        "max": 1,
        "mean": 1,
        "median": 1,
        "stdDev": 0,
        "timeSeriesEligible": false,
        "timeSeriesEligibilityReason": "notADate",
        "timeStep": null,
        "timeUnit": null,
        "targetLeakage": "SKIPPED_DETECTION"
    }
    """


@pytest.fixture
def feature_json_no_leakage():
    return """
    {
        "id": 34,
        "name": "Claim_Amount",
        "featureType": "Numeric",
        "importance": 1,
        "lowInformation": false,
        "uniqueCount": 200,
        "naCount": 0,
        "dateFormat": null,
        "projectId": "56cdd3baccf94e69733ff20f",
        "min": 1,
        "max": 1,
        "mean": 1,
        "median": 1,
        "stdDev": 0,
        "timeSeriesEligible": false,
        "timeSeriesEligibilityReason": "notADate",
        "timeStep": null,
        "timeUnit": null,
        "targetLeakage": "FALSE"
    }
    """


@pytest.fixture
def feature_json_leakage():
    return """
    {
        "id": 34,
        "name": "Claim_Amount",
        "featureType": "Numeric",
        "importance": 1,
        "lowInformation": true,
        "uniqueCount": 200,
        "naCount": 0,
        "dateFormat": null,
        "projectId": "56cdd3baccf94e69733ff20f",
        "min": 1,
        "max": 1,
        "mean": 1,
        "median": 1,
        "stdDev": 0,
        "timeSeriesEligible": false,
        "timeSeriesEligibilityReason": "notADate",
        "timeStep": null,
        "timeUnit": null,
        "targetLeakage": "HIGH_RISK"
    }
    """


@pytest.fixture
def feature_json_invalid_leakage():
    return """
    {
        "id": 34,
        "name": "Claim_Amount",
        "featureType": "Numeric",
        "importance": 1,
        "lowInformation": true,
        "uniqueCount": 200,
        "naCount": 0,
        "dateFormat": null,
        "projectId": "56cdd3baccf94e69733ff20f",
        "min": 1,
        "max": 1,
        "mean": 1,
        "median": 1,
        "stdDev": 0,
        "timeSeriesEligible": false,
        "timeSeriesEligibilityReason": "notADate",
        "timeStep": null,
        "timeUnit": null,
        "targetLeakage": 4
    }
    """


@pytest.fixture
def modeling_feature_server_data(project_id):
    return {
        'featureType': 'Numeric',
        'lowInformation': False,
        'uniqueCount': 252,
        'name': 'a_feat',
        'min': 0.052,
        'importance': 0.0039430549403626225,
        'max': 0.15,
        'dateFormat': None,
        'median': 0.091,
        'stdDev': 0.014,
        'projectId': project_id,
        'naCount': 0,
        'parentFeatureNames': [
            'original'
        ],
        'mean': 0.092
    }


@pytest.fixture
def modeling_features_list_server_data(modeling_feature_server_data):
    return {
        'count': 1,
        'next': None,
        'previous': None,
        'data': [modeling_feature_server_data]
    }


@pytest.fixture
def modeling_features_with_next_page_server_data(modeling_feature_server_data, project_url):
    next_page_url = '{}modelingFeatures/?offset=1&limit=1'.format(project_url)
    feat = dict(modeling_feature_server_data)
    feat['name'] = 'first_page'
    return {
        'count': 1,
        'next': next_page_url,
        'previous': None,
        'data': [feat]
    }


@pytest.fixture
def modeling_features_with_previous_page_server_data(modeling_feature_server_data, project_url):
    previous_page_url = '{}modelingFeatures/?offset=0&limit=1'.format(project_url)
    feat = dict(modeling_feature_server_data)
    feat['name'] = 'second_page'
    return {
        'count': 1,
        'next': None,
        'previous': previous_page_url,
        'data': [feat]
    }


@pytest.fixture
def feature_sans_importance_json():
    return """
    {
        "id": 34,
        "name": "Claim_Amount",
        "featureType": "Numeric",
        "importance": null,
        "lowInformation": false,
        "uniqueCount": 200,
        "naCount": 0,
        "dateFormat": null,
        "projectId": "556cdfbb100d2b0e88585195",
        "min": 1,
        "max": 1,
        "mean": 1,
        "median": 1,
        "stdDev": 0,
        "timeSeriesEligible": false,
        "timeSeriesEligibilityReason": "notADate",
        "timeStep": null,
        "timeUnit": null,
        "targetLeakage": "SKIPPED_DETECTION"
    }
    """


@pytest.fixture
def claim_amount_json():
    return """
    {
        "id": 34,
        "name": "Claim Amount",
        "featureType": "Numeric",
        "importance": null,
        "lowInformation": false,
        "uniqueCount": 200,
        "naCount": 0,
        "dateFormat": null,
        "projectId": "556cdfbb100d2b0e88585195",
        "min": 1,
        "max": 1,
        "mean": 1,
        "median": 1,
        "stdDev": 0,
        "timeSeriesEligible": false,
        "timeSeriesEligibilityReason": "notADate",
        "timeStep": null,
        "timeUnit": null,
        "targetLeakage": "SKIPPED_DETECTION"
    }
    """


@pytest.fixture
def low_info_feature_json():
    """The information for a feature that has been marked as low info

    DataRobot does not finish gathering the naCount or generate a featureType
    in these situations
    """
    return """
    {
    "featureType": null,
    "lowInformation": true,
    "name": "mths_since_last_major_derog",
    "uniqueCount": 0,
    "importance": null,
    "id": 25,
    "naCount": null,
    "dateFormat": null,
    "projectId": "556cdfbb100d2b0e88585195",
    "min": null,
    "max": null,
    "mean": null,
    "median": null,
    "stdDev": null,
    "timeSeriesEligible": false,
    "timeSeriesEligibilityReason": "notADate",
    "timeStep": null,
    "timeUnit": null,
    "targetLeakage": "SKIPPED_DETECTION"
    }
    """


@pytest.fixture
def float_feature_json():
    return """
    {
        "uniqueCount": 1,
        "lowInformation": false,
        "name": "float",
        "dateFormat": null,
        "importance": 0.2648777523946366,
        "projectId": "556cdfbb100d2b0e88585195",
        "featureType": "Date",
        "id": 10,
        "naCount": 0,
        "min": 123.45,
        "max": 123.45,
        "mean": 123.45,
        "median": 123.45,
        "stdDev": 0,
        "timeSeriesEligible": false,
        "timeSeriesEligibilityReason": "notADate",
        "timeStep": null,
        "timeUnit": null,
        "targetLeakage": "SKIPPED_DETECTION"
    }
    """


@pytest.fixture
def date_feature_json():
    return """
    {
        "uniqueCount": 157,
        "lowInformation": false,
        "name": "dates",
        "dateFormat": "%Y-%m-%d %H:%M:%S",
        "importance": 0.2648777523946366,
        "projectId": "556cdfbb100d2b0e88585195",
        "featureType": "Date",
        "id": 1,
        "naCount": 0,
        "min": "2017-03-04T00:00:00",
        "max": "2017-06-15T04:40:39",
        "mean": "2017-04-24T14:20:19",
        "median": "2017-04-24T14:20:19",
        "stdDev": "30.089 days",
        "timeSeriesEligible": true,
        "timeSeriesEligibilityReason": "suitable",
        "timeStep": 3,
        "timeUnit": "MONTH",
        "targetLeakage": "SKIPPED_DETECTION"
    }
    """


@pytest.fixture
def feature_server_data(feature_json):
    return json.loads(feature_json)


def test_future_proof(feature_server_data):
    Feature.from_server_data(dict(feature_server_data, future='key'))


@responses.activate
@pytest.mark.usefixtures('client')
def test_features(list_features_json):
    responses.add(responses.GET,
                  'https://host_name.com/projects/p-id/features/',
                  body=list_features_json)
    feature, = Project('p-id').get_features()
    assert isinstance(feature, Feature)
    assert feature.id == 34
    assert feature.name == 'Claim_Amount'
    assert feature.feature_type == 'Numeric'
    assert feature.importance == 1
    assert feature.low_information is False
    assert feature.unique_count == 200
    assert feature.na_count == 0
    assert feature.project_id == '556cdfbb100d2b0e88585195'
    assert feature.min == 1
    assert feature.max == 1
    assert feature.median == 1
    assert feature.median == 1
    assert feature.std_dev == 0
    assert feature.time_series_eligible is False
    assert feature.time_series_eligibility_reason == 'notADate'
    assert feature.time_step is None
    assert feature.time_unit is None


@responses.activate
@pytest.mark.usefixtures('client')
def test_feature(feature_sans_importance_json):
    responses.add(responses.GET,
                  'https://host_name.com/projects/p-id/features/f-id/',
                  body=feature_sans_importance_json)
    feature = Feature.get('p-id', 'f-id')
    assert isinstance(feature, Feature)
    assert feature.id == 34
    assert feature.name == 'Claim_Amount'
    assert feature.feature_type == 'Numeric'
    assert feature.importance is None
    assert feature.low_information is False
    assert feature.unique_count == 200
    assert feature.na_count == 0
    assert feature.date_format is None
    assert feature.project_id == '556cdfbb100d2b0e88585195'
    assert feature.min == 1
    assert feature.max == 1
    assert feature.median == 1
    assert feature.median == 1
    assert feature.std_dev == 0
    assert feature.target_leakage == "SKIPPED_DETECTION"


@responses.activate
@pytest.mark.usefixtures('client')
def test_feature_no_leakage(feature_json_no_leakage):
    responses.add(responses.GET,
                  'https://host_name.com/projects/p-id/features/f-id/',
                  body=feature_json_no_leakage)
    feature = Feature.get('p-id', 'f-id')
    assert isinstance(feature, Feature)
    assert feature.low_information is False
    assert feature.target_leakage == "FALSE"


@responses.activate
@pytest.mark.usefixtures('client')
def test_feature_leakage(feature_json_leakage):
    responses.add(responses.GET,
                  'https://host_name.com/projects/p-id/features/f-id/',
                  body=feature_json_leakage)
    feature = Feature.get('p-id', 'f-id')
    assert isinstance(feature, Feature)
    assert feature.low_information is True
    assert feature.target_leakage == "HIGH_RISK"


@responses.activate
@pytest.mark.usefixtures('client')
def test_feature_invalid_leakage(feature_json_invalid_leakage):
    responses.add(responses.GET,
                  'https://host_name.com/projects/p-id/features/f-id/',
                  body=feature_json_invalid_leakage)
    with pytest.raises(DataError):
        Feature.get('p-id', 'f-id')


@responses.activate
@pytest.mark.usefixtures('client')
def test_feature_with_low_info(low_info_feature_json):
    responses.add(responses.GET,
                  'https://host_name.com/projects/p-id/features/f-id/',
                  body=low_info_feature_json)
    feature = Feature.get('p-id', 'f-id')
    assert feature.na_count is None
    assert feature.feature_type is None
    assert feature.min is None
    assert feature.max is None
    assert feature.median is None
    assert feature.median is None
    assert feature.std_dev is None


@responses.activate
@pytest.mark.usefixtures('client')
def test_feature_with_space(claim_amount_json):
    responses.add(responses.GET,
                  'https://host_name.com/projects/p-id/features/Claim%20Amount/',
                  body=claim_amount_json)
    feature = Feature.get('p-id', 'Claim Amount')
    assert isinstance(feature, Feature)


@responses.activate
@pytest.mark.usefixtures('client')
def test_float_feature(float_feature_json):
    responses.add(responses.GET, 'https://host_name.com/projects/p-id/features/float/',
                  body=float_feature_json)
    feature = Feature.get('p-id', 'float')
    assert isinstance(feature, Feature)
    assert feature.min == 123.45
    assert feature.max == 123.45
    assert feature.mean == 123.45
    assert feature.median == 123.45
    assert feature.std_dev == 0


@responses.activate
@pytest.mark.usefixtures('client')
def test_date_feature(date_feature_json):
    responses.add(responses.GET, 'https://host_name.com/projects/p-id/features/dates/',
                  body=date_feature_json)
    feature = Feature.get('p-id', 'dates')
    assert isinstance(feature, Feature)
    assert feature.date_format == json.loads(date_feature_json)['dateFormat']
    assert feature.min == "2017-03-04T00:00:00"
    assert feature.max == "2017-06-15T04:40:39"
    assert feature.mean == "2017-04-24T14:20:19"
    assert feature.median == "2017-04-24T14:20:19"
    assert feature.std_dev == "30.089 days"
    assert feature.time_series_eligible is True
    assert feature.time_series_eligibility_reason == 'suitable'
    assert feature.time_step == 3
    assert feature.time_unit == 'MONTH'


@responses.activate
@pytest.mark.usefixtures('client')
def test_get_multiseries_properties_pre_computed(date_feature_json,
                                                 multiseries_precomputed_properties):
    feature = Feature.from_server_data(json.loads(date_feature_json))

    properties = feature.get_multiseries_properties(['series_id'])
    assert properties == {'time_series_eligible': True,
                          'time_step': multiseries_precomputed_properties['timeStep'],
                          'time_unit': multiseries_precomputed_properties['timeUnit']}
    assert len(responses.calls) == 1


@responses.activate
@pytest.mark.usefixtures('client')
def test_get_multiseries_properties_with_computation(date_feature_json,
                                                     multiseries_postcomputed_properties):
    feature = Feature.from_server_data(json.loads(date_feature_json))

    properties = feature.get_multiseries_properties(['series_id'])
    assert properties == {'time_series_eligible': True,
                          'time_step': multiseries_postcomputed_properties['timeStep'],
                          'time_unit': multiseries_postcomputed_properties['timeUnit']}
    assert len(responses.calls) == 4


@responses.activate
@pytest.mark.usefixtures('multiseries_ineligible_properties', 'client')
def test_get_multiseries_properties_not_eligible(float_feature_json):
    feature = Feature.from_server_data(json.loads(float_feature_json))

    properties = feature.get_multiseries_properties(['series_id'])
    assert properties == {'time_series_eligible': False, 'time_step': None, 'time_unit': None}
    assert len(responses.calls) == 4


def test_feature_with_non_ascii_name(feature_server_data, unicode_string):
    data_copy = dict(feature_server_data)
    data_copy['name'] = unicode_string
    feature = Feature.from_server_data(data_copy)
    print(feature)  # actually part of the test - this used to fail (testing __repr__)


@pytest.mark.usefixtures('known_warning')
def test_bc_instantiate_feature_from_dict(feature_server_data):
    feature = Feature(feature_server_data)
    assert feature.id == feature_server_data['id']
    assert feature.name == feature_server_data['name']
    assert feature.feature_type == feature_server_data['featureType']
    assert feature.importance == feature_server_data['importance']
    assert feature.low_information == feature_server_data['lowInformation']
    assert feature.unique_count == feature_server_data['uniqueCount']
    assert feature.na_count == feature_server_data['naCount']
    assert feature.project_id == '556cdfbb100d2b0e88585195'


def test_modeling_feature_future_proof(modeling_feature_server_data):
    future_data = dict(modeling_feature_server_data, new='new')
    ModelingFeature.from_server_data(future_data)


@responses.activate
@pytest.mark.usefixtures('client')
def test_get_modeling_feature(modeling_feature_server_data, project_url, project_id):
    name = modeling_feature_server_data['name']
    responses.add(responses.GET, '{}modelingFeatures/a_feat/'.format(project_url),
                  json=modeling_feature_server_data)

    feature = ModelingFeature.get(project_id, name)
    assert feature.project_id == project_id
    assert feature.name == modeling_feature_server_data['name']
    assert feature.feature_type == modeling_feature_server_data['featureType']
    assert feature.importance == modeling_feature_server_data['importance']
    assert feature.low_information == modeling_feature_server_data['lowInformation']
    assert feature.unique_count == modeling_feature_server_data['uniqueCount']
    assert feature.na_count == modeling_feature_server_data['naCount']
    assert feature.date_format == modeling_feature_server_data['dateFormat']
    assert feature.min == modeling_feature_server_data['min']
    assert feature.max == modeling_feature_server_data['max']
    assert feature.mean == modeling_feature_server_data['mean']
    assert feature.median == modeling_feature_server_data['median']
    assert feature.std_dev == modeling_feature_server_data['stdDev']
    assert feature.parent_feature_names == modeling_feature_server_data['parentFeatureNames']


@responses.activate
@pytest.mark.usefixtures('client')
def test_list_modeling_features(modeling_features_list_server_data, project_url, project):
    responses.add(responses.GET, '{}modelingFeatures/'.format(project_url),
                  json=modeling_features_list_server_data)
    feats = project.get_modeling_features()

    assert len(feats) == len(modeling_features_list_server_data['data'])
    assert isinstance(feats[0], ModelingFeature)
    assert feats[0].name == modeling_features_list_server_data['data'][0]['name']


@responses.activate
@pytest.mark.usefixtures('client')
def test_list_modeling_features_paginated(modeling_features_with_next_page_server_data,
                                          modeling_features_with_previous_page_server_data,
                                          project, project_url):
    with responses.RequestsMock() as rsps:
        rsps.add(responses.GET, '{}modelingFeatures/'.format(project_url),
                 json=modeling_features_with_next_page_server_data)
        rsps.add(responses.GET, '{}modelingFeatures/'.format(project_url),
                 json=modeling_features_with_previous_page_server_data)
        feats = project.get_modeling_features(batch_size=1)

        first_page_req = rsps.calls[0].request
        second_page_req = rsps.calls[1].request
        assert {'limit': ['1']} == parse_qs(urlparse(first_page_req.url).query)
        assert {'limit': ['1'], 'offset': ['1']} == parse_qs(urlparse(second_page_req.url).query)

    assert len(feats) == 2
    names = {'first_page', 'second_page'}
    assert {feat.name for feat in feats} == names



"""This module is not considered part of the public interface. As of 2.3, anything here
may change or be removed without warning."""

import platform
import os

import requests
# Import the urllib3.Retry class from requests.adapters rather than urllib3
# directly because of the mess of things requests made by vendoring the
# package. This is the best way to be sure we get the correct class, otherwise
# _weird_ things will happen.
from requests.adapters import HTTPAdapter, Retry
import six
import trafaret as t
from requests_toolbelt.multipart.encoder import MultipartEncoder
from six.moves.urllib_parse import urlparse

from .enums import DEFAULT_TIMEOUT
from . import __version__
from .utils import to_api
from . import errors


class RESTClientObject(requests.Session):
    """
    Parameters
        connect_timeout
            timeout for http request and connection
        headers
            headers for outgoing requests
    """
    @classmethod
    def from_config(cls, config):
        return cls(auth=config.token, endpoint=config.endpoint,
                   connect_timeout=config.connect_timeout, verify=config.ssl_verify,
                   user_agent_suffix=config.user_agent_suffix, max_retries=config.max_retries)

    def __init__(self, auth=None, endpoint=None, connect_timeout=DEFAULT_TIMEOUT.CONNECT,
                 verify=True, user_agent_suffix=None, max_retries=None):
        super(RESTClientObject, self).__init__()
        # Note: As of 2.3, `endpoint` is required
        self.endpoint = endpoint
        self.domain = '{}://{}'.format(urlparse(self.endpoint).scheme,
                                       urlparse(self.endpoint).netloc)
        self.token = auth
        if connect_timeout is None:
            connect_timeout = DEFAULT_TIMEOUT.CONNECT
        self.connect_timeout = connect_timeout
        self.user_agent_header = self._make_user_agent_header(suffix=user_agent_suffix)
        self.headers.update(self.user_agent_header)
        self.token_header = {'Authorization': 'Token {}'.format(self.token)}
        self.headers.update(self.token_header)
        self.verify = verify
        if max_retries is None:
            # by default, retry connect error but not read error for _all_ requests
            max_retries = Retry(connect=5, read=0, method_whitelist=False, backoff_factor=0.1)
        self.mount('http://', HTTPAdapter(max_retries=max_retries))
        self.mount('https://', HTTPAdapter(max_retries=max_retries))

    @staticmethod
    def _make_user_agent_header(suffix=None):
        agent_components = []
        agent_components.append('DataRobotPythonClient/{}'.format(__version__))
        agent_components.append('({} {} {})'.format(platform.system(),
                                                    platform.release(),
                                                    platform.machine()))
        if suffix:
            agent_components.append(suffix)
        return {'User-Agent': ' '.join(agent_components)}

    def _join_endpoint(self, url):
        return '{}/{}'.format(self.endpoint, url)

    def strip_endpoint(self, url):
        trailing = '' if self.endpoint.endswith('/') else '/'
        expected = '{}{}'.format(self.endpoint, trailing)
        if not url.startswith(expected):
            raise ValueError('unexpected url format: {} does not start with {}'.format(url,
                                                                                       expected))
        return url.split(expected)[1]

    def request(self, method, url, join_endpoint=False, **kwargs):
        kwargs.setdefault('timeout', (self.connect_timeout, DEFAULT_TIMEOUT.READ))
        if not url.startswith('http') or join_endpoint:
            url = self._join_endpoint(url)
        response = super(RESTClientObject, self).request(method, url, **kwargs)
        if not response:
            handle_http_error(response, **kwargs)
        return response

    def get(self, url, params=None, **kwargs):
        return self.request('get', url, params=to_api(params), **kwargs)

    def post(self, url, data=None, keep_attrs=None, **kwargs):
        if data:
            kwargs['json'] = to_api(data, keep_attrs)
        return self.request('post', url, **kwargs)

    def patch(self, url, data=None, **kwargs):
        if data:
            kwargs['json'] = to_api(data)
        return self.request('patch', url, **kwargs)

    def build_request_with_file(self, method, url,
                                fname,
                                form_data=None,
                                content=None,
                                file_path=None,
                                filelike=None,
                                read_timeout=DEFAULT_TIMEOUT.READ,
                                file_field_name='file'):
        """Build request with a file that will use special
        MultipartEncoder instance (lazy load file).


        This method supports uploading a file on local disk, string content,
        or a file-like descriptor. ``fname`` is a required parameter, and
        only one of the other three parameters can be provided.

        Parameters
        ----------
        method : str.
            Method of request. This parameter is required, it can be
            'POST' or 'PUT' either 'PATCH'.
        url : str.
            Url that will be used it this request.
        fname : name of file
            This parameter is required, even when providing a file-like object
            or string content.
        content : str
            The content buffer of the file you would like to upload.
        file_path : str
            The path to a file on a local file system.
        filelike : file-like
            An open file descriptor to a file.
        read_timeout : float
            The number of seconds to wait after the server receives the file that we are
            willing to wait for the beginning of a response. Large file uploads may take
            significant time.
        file_field_name : str
            The name of the form field we will put the data into (Defaults to 'file').

        Returns
        -------
        response : response object.

        """

        bad_args_msg = ('Upload should be used either with content buffer '
                        'or with path to file on local filesystem or with '
                        'open file descriptor')
        assert sum((bool(content),
                    bool(file_path),
                    bool(filelike))) == 1, bad_args_msg

        if file_path:
            if not os.path.exists(file_path):
                raise ValueError(
                    u'Provided file does not exist {}'.format(file_path))
            fields = {file_field_name: (fname, open(file_path, 'rb'))}

        elif filelike:
            filelike.seek(0)
            fields = {file_field_name: (fname, filelike)}
        else:
            if not isinstance(content, six.binary_type):
                raise AssertionError('bytes type required in content')
            fields = {file_field_name: (fname, content)}

        form_data = form_data or {}
        data_for_encoder = to_api(form_data)
        data_for_encoder.update(fields)

        encoder = MultipartEncoder(fields=data_for_encoder)
        headers = {'Content-Type': encoder.content_type}
        return self.request(method, url, headers=headers, data=encoder,
                            timeout=(self.connect_timeout, read_timeout))


def _http_message(response):
    if response.status_code == 401:
        message = ('The server is saying you are not properly '
                   'authenticated. Please make sure your API '
                   'token is valid.')
    elif response.headers['content-type'] == 'application/json':
        message = response.json()
    else:
        message = response.content.decode('ascii')[:79]
    return message


def handle_http_error(response, **kwargs):
    message = _http_message(response)
    if 400 <= response.status_code < 500:
        exception_type = errors.ClientError
        # One-off approach to raising special exception for now. We'll do something more
        # systematic when we have more of these:
        try:
            parsed_json = response.json()  # May raise ValueError if response isn't JSON
            if parsed_json.get('errorName') == 'JobAlreadyAdded':
                exception_type = errors.JobAlreadyRequested
        except ValueError:
            parsed_json = {}
        template = '{} client error: {}'
        exc_message = template.format(response.status_code, message)
        raise exception_type(exc_message, response.status_code, json=parsed_json)
    else:
        template = '{} server error: {}'
        exc_message = template.format(response.status_code, message)
        raise errors.ServerError(exc_message, response.status_code)


class DataRobotClientConfig(object):
    """
    This class contains all of the client configuration variables that are known to
    the DataRobot client.

    Values are allowed to be None in this object. The __init__ of RESTClientObject will
    provide any defaults that should be applied if the user does not specify in the config
    """
    _converter = t.Dict({
        t.Key('endpoint'): t.String(),
        t.Key('token'): t.String(),
        t.Key('connect_timeout', optional=True): t.Int(),
        t.Key('ssl_verify', optional=True): t.Or(t.Bool(), t.String()),
        t.Key('user_agent_suffix', optional=True): t.String(),
        t.Key('max_retries', optional=True): t.Int()
    }).allow_extra('*')
    _fields = {k.to_name or k.name for k in _converter.keys}

    def __init__(self, endpoint, token, connect_timeout=None, ssl_verify=True,
                 user_agent_suffix=None, max_retries=None):
        self.endpoint = endpoint
        self.token = token
        self.connect_timeout = connect_timeout
        self.ssl_verify = ssl_verify
        self.user_agent_suffix = user_agent_suffix
        self.max_retries = max_retries

    @classmethod
    def from_data(cls, data):
        checked = {k: v
                   for k, v in cls._converter.check(data).items()
                   if k in cls._fields}
        return cls(**checked)



import pytest
import responses
import json
from datarobot import errors


@pytest.fixture
def cv_scores_url(project_id, model_id):
    return 'https://host_name.com/projects/{}/models/{}/crossValidationScores/'.format(project_id,
                                                                                       model_id)


@pytest.fixture
def cv_scores():
    return {"metric1": {"1.0": "val", "2.0": "val2"},
            "metric2": {"1.0": "val3", "2.0": "val4"},
            }


@pytest.fixture
def cv_scores_get_response(cv_scores_url, cv_scores):
    responses.add(
        responses.GET,
        cv_scores_url,
        status=200,
        content_type='applicaiton/json',
        body=json.dumps(cv_scores)
    )


@pytest.fixture
def cv_scores_get_response_404(cv_scores_url, cv_scores):
    responses.add(
        responses.GET,
        cv_scores_url,
        status=404,
        content_type='applicaiton/json',
        body="object does not contain cross validation scores"
    )


@pytest.fixture
def cv_scores_get_response_filter_by_metric(cv_scores_url, cv_scores):
    url = cv_scores_url + "?metric=metric1"
    responses.add(
        responses.GET,
        url,
        status=200,
        content_type='applicaiton/json',
        body=json.dumps(cv_scores)
    )


@pytest.fixture
def cv_scores_get_response_filter_by_partition(cv_scores_url, cv_scores):
    url = cv_scores_url + "?partition=2"
    responses.add(
        responses.GET,
        url,
        status=200,
        content_type='applicaiton/json',
        body=json.dumps(cv_scores)
    )


@pytest.fixture
def cv_scores_get_response_filter_by_partition_and_metric(cv_scores_url, cv_scores):
    url = cv_scores_url + "?partition=2&metric=metric1"
    responses.add(
        responses.GET,
        url,
        status=200,
        content_type='applicaiton/json',
        body=json.dumps(cv_scores)
    )


@responses.activate
@pytest.mark.usefixtures('client',
                         'cv_scores_get_response'
                         )
def test_get_cross_validation_scores(one_model, cv_scores):
    cv_scores_retrieved = one_model.get_cross_validation_scores()
    assert cv_scores_retrieved == cv_scores


@responses.activate
@pytest.mark.usefixtures('client',
                         'cv_scores_get_response_filter_by_metric'
                         )
def test_get_cross_validation_scores_filtered_by_metric(one_model, cv_scores):
    cv_scores_retrieved = one_model.get_cross_validation_scores(metric='metric1')
    assert cv_scores_retrieved == cv_scores


@responses.activate
@pytest.mark.usefixtures('client',
                         'cv_scores_get_response_filter_by_partition'
                         )
def test_get_cross_validation_scores_filtered_by_partition(one_model, cv_scores):
    cv_scores_retrieved = one_model.get_cross_validation_scores(partition=2)
    assert cv_scores_retrieved == cv_scores


@responses.activate
@pytest.mark.usefixtures('client',
                         'cv_scores_get_response_filter_by_partition_and_metric'
                         )
def test_get_cross_validation_scores_filtered_by_partition_and_metric(one_model, cv_scores):
    cv_scores_retrieved = one_model.get_cross_validation_scores(partition=2, metric='metric1')
    assert cv_scores_retrieved == cv_scores


@responses.activate
@pytest.mark.usefixtures('client',
                         'cv_scores_get_response_404'
                         )
def test_get_cross_validation_scores_404(one_model, cv_scores):
    with pytest.raises(errors.ClientError):
        one_model.get_cross_validation_scores()



from collections import namedtuple
import itertools

import pandas as pd
from six.moves.urllib_parse import urlparse, parse_qs
import trafaret as t

from datarobot import errors, enums
from ..utils import from_api
from .api_object import APIObject


class RowsIterator(object):
    """
    Yields all available records one-by-one.

    While iterating, fetches rows from API with a series of requests with specified limit
    Stops iterating when API has no more objects to fetch
    """
    _trafaret = t.Dict({
        t.Key('data'): t.List(t.Dict().allow_extra('*')),
        t.Key('next'): t.String | t.Null,
    }).allow_extra('*')

    def __init__(self, client, path, limit=None):
        self.client = client
        self.path = path
        self.query_params = dict(limit=limit)

        self._queue = []
        self._stop_flag = None

    def __iter__(self):
        return self

    def next(self):
        if self._stop_flag and not self._queue:
            raise StopIteration()

        if not self._queue:
            items, next_url = self._load_more()
            self._queue.extend(items)
            if not self._queue:
                raise StopIteration()

            self._stop_flag = next_url is None
            self._stop_flag |= self.is_exhausted(
                len(items),
                self.query_params.get('limit'),
            )
            if next_url is not None:
                self._save_next_url(next_url)

        data_dict = self._queue.pop(0)
        return data_dict

    __next__ = next

    @staticmethod
    def is_exhausted(loaded_count, limit):
        if limit is None:
            # If there was no limit param, assume all items were fetched
            return True

        # If we asked for 20 items but got 10 items that means our collection is exhausted
        return loaded_count < limit

    def _load_more(self):
        """Fetch more items from API, validate response, save next url"""
        response = self.client.get(self.path, params=self.query_params)
        if response.status_code != 200:
            e_msg = 'Server returned unexpected status code'
            raise errors.ServerError(
                e_msg.format(response.status_code),
                response.status_code,
            )

        body = self._trafaret.check(response.json())
        return body['data'], body['next']

    def _save_next_url(self, next_url):
        """Next URL should contain `limit` & `offset` query params"""
        parsed = urlparse(next_url)
        parsed_qs = parse_qs(parsed.query)
        self.query_params = {
            key: int(values_list[0])
            for key, values_list in parsed_qs.items()
        }


TrainingPredictionsRow = namedtuple(
    'TrainingPredictionsRow',
    (
        'row_id,partition_id,prediction,prediction_values,'
        'timestamp,forecast_point,forecast_distance,series_id'
    )
)


class TrainingPredictionsIterator(RowsIterator):
    """
    Lazily fetches training predictions from DataRobot API in chunks of specified size and then
    iterates rows from responses as named tuples. Each row represents a training prediction
    computed for a dataset's row. Each named tuple has the following structure:

    Attributes
    ----------
    row_id : int
        id of the record in original dataset for which training prediction is calculated
    partition_id : str or float
        id of the data partition that the row belongs to
    prediction : float
        the model's prediction for this data row
    prediction_values : list of dictionaries
        an array of dictionaries with a schema described as ``PredictionValue``
    timestamp : str or None
        (New in version v2.11) an ISO string representing the time of the prediction
        in time series project; may be None for non-time series projects
    forecast_point : str or None
        (New in version v2.11) an ISO string representing the point in time
        used as a basis to generate the predictions in time series project;
        may be None for non-time series projects
    forecast_distance : str or None
        (New in version v2.11) how many time steps are between the forecast point and the
        timestamp in time series project; None for non-time series projects
    series_id : str or None
        (New in version v2.11) the id of the series in a multiseries project;
        may be NaN for single series projects; None for non-time series projects

    Notes
    -----
    Each ``PredictionValue`` dict contains these keys:

        label
            describes what this model output corresponds to. For regression
            projects, it is the name of the target feature. For classification and multiclass
            projects, it is a label from the target feature.
        value
            the output of the prediction. For regression projects, it is the
            predicted value of the target. For classification and multiclass projects, it is
            the predicted probability that the row belongs to the class identified by the label.

    Examples
    --------
    .. code-block:: python

        import datarobot as dr

        # Fetch existing training predictions by their id
        training_predictions = dr.TrainingPredictions.get(project_id, prediction_id)

        # Iterate over predictions
        for row in training_predictions.iterate_rows()
            print(row.row_id, row.prediction)
    """
    _row_trafaret = t.Dict({
        t.Key('row_id'): t.Int(),
        t.Key('partition_id'): t.String() | t.Float(),
        t.Key('prediction'): t.Float() | t.String(),
        t.Key('prediction_values'): t.List(t.Dict({
            t.Key('label'): t.Float | t.String,
            t.Key('value'): t.Float,
        }).ignore_extra('*')),
        t.Key('timestamp', default=None): t.String | t.Null,
        t.Key('forecast_point', default=None): t.String | t.Null,
        t.Key('forecast_distance', default=None): t.Int | t.Null,
        t.Key('series_id', default=None): t.String | t.Null,
    }).ignore_extra('*')

    def next(self):
        row_dict = super(TrainingPredictionsIterator, self).next()
        row_dict = self._row_trafaret.check(row_dict)
        return TrainingPredictionsRow(**row_dict)

    __next__ = next

    def _load_more(self):
        rows, next_url = super(TrainingPredictionsIterator, self)._load_more()
        return from_api(rows), next_url


class TrainingPredictions(APIObject):
    """
    Represents training predictions metadata and provides access to prediction results.

    Attributes
    ----------
    project_id : str
        id of the project the model belongs to
    model_id : str
        id of the model
    prediction_id : str
        id of generated predictions

    Examples
    --------
    Compute training predictions for a model on the whole dataset

    .. code-block:: python

        import datarobot as dr

        # Request calculation of training predictions
        training_predictions_job = model.request_training_predictions(dr.enums.DATA_SUBSET.ALL)
        training_predictions = training_predictions_job.get_result_when_complete()
        print('Training predictions {} are ready'.format(training_predictions.prediction_id))

        # Iterate over actual predictions
        for row in training_predictions.iterate_rows():
            print(row.row_id, row.partition_id, row.prediction)

    List all training predictions for a project

    .. code-block:: python

        import datarobot as dr

        # Fetch all training predictions for a project
        all_training_predictions = dr.TrainingPredictions.list(project_id)

        # Inspect all calculated training predictions
        for training_predictions in all_training_predictions:
            print(
                'Prediction {} is made for data subset "{}"'.format(
                    training_predictions.prediction_id,
                    training_predictions.data_subset,
                )
            )

    Retrieve training predictions by id

    .. code-block:: python

        import datarobot as dr

        # Getting training predictions by id
        training_predictions = dr.TrainingPredictions.get(project_id, prediction_id)

        # Iterate over actual predictions
        for row in training_predictions.iterate_rows():
            print(row.row_id, row.partition_id, row.prediction)
    """

    def __init__(self, project_id, prediction_id, model_id=None, data_subset=None):
        self.project_id = project_id
        self.model_id = model_id
        self.prediction_id = prediction_id
        self.path = self.build_path(project_id, prediction_id)
        self.data_subset = data_subset

    @staticmethod
    def build_path(project_id, prediction_id=None):
        if prediction_id is not None:
            return 'projects/{}/trainingPredictions/{}/'.format(project_id, prediction_id)

        return 'projects/{}/trainingPredictions/'.format(project_id)

    @classmethod
    def from_location(cls, location, data_subset=None, model_id=None):
        head, tail = location.split('/trainingPredictions/', 1)
        project_id, prediction_id = head.split('/')[-1], tail.split('/')[0]
        return cls(project_id, prediction_id, data_subset=data_subset, model_id=model_id)

    @classmethod
    def list(cls, project_id):
        """
        Fetch all the computed training predictions for a project.

        Parameters
        ----------
        project_id : str
            id of the project

        Returns
        -------
        A list of :py:class:`TrainingPredictions` objects
        """
        _trafaret = t.Dict({
            t.Key('data'): t.List(t.Dict({
                t.Key('url'): t.String(),
                t.Key('model_id'): t.String(),
                t.Key('data_subset'): t.String(),
            }).ignore_extra('*')),
        }).ignore_extra('*')

        path = cls.build_path(project_id)
        converted = from_api(cls._server_data(path))
        validated = _trafaret.check(converted)['data']
        return [
            cls.from_location(
                item['url'],
                data_subset=item['data_subset'],
                model_id=item['model_id'],
            ) for item in validated
        ]

    @classmethod
    def get(cls, project_id, prediction_id):
        """
        Retrieve training predictions on a specified data set.

        Parameters
        ----------
        project_id : str
            id of the project the model belongs to
        prediction_id : str
            id of the prediction set

        Returns
        -------
        :py:class:`TrainingPredictions` object which is ready to operate with specified predictions
        """
        return cls(project_id, prediction_id)

    def iterate_rows(self, batch_size=None):
        """
        Retrieve training prediction rows as an iterator.

        Parameters
        ----------
        batch_size : int, optional
            maximum number of training prediction rows to fetch per request

        Returns
        -------
        iterator : :py:class:`TrainingPredictionsIterator`
            an iterator which yields named tuples representing training prediction rows
        """
        return TrainingPredictionsIterator(
            self._client,
            self.path,
            limit=batch_size,
        )

    def get_all_as_dataframe(self, class_prefix=enums.PREDICTION_PREFIX.DEFAULT):
        """
        Retrieve all training prediction rows and return them as a pandas.DataFrame.

        Returned dataframe has the following structure:
            - row_id : row id from the original dataset
            - prediction : the model's prediction for this row
            - class_<label> : the probability that the target is this class (only appears for
              classification and multiclass projects)
            - timestamp : the time of the prediction (only appears for time series projects)
            - forecast_point : the point in time used as a basis to generate the predictions
              (only appears for time series projects)
            - forecast_distance : how many time steps are between timestamp and forecast_point
              (only appears for time series projects)
            - series_id : he id of the series in a multiseries project
              or None for a single series project
              (only appears for time series projects)

        Parameters
        ----------
        class_prefix : str, optional
            The prefix to append to labels in the final dataframe. Default is ``class_``
            (e.g., apple -> class_apple)

        Returns
        -------
        dataframe: pandas.DataFrame
        """
        rows = self.iterate_rows()

        tmp, rows = itertools.tee(rows)
        first_row = next(tmp, None)
        is_classification = first_row is not None and len(first_row.prediction_values) > 1

        if is_classification:
            labels = (p['label'] for p in first_row.prediction_values)
            columns = self._get_classification_columns(labels, class_prefix=class_prefix)
            return self._build_classification_dataframe(rows, columns)
        elif self._is_time_series_project:
            return self._build_timeseries_dataframe(rows)
        else:
            return self._build_regression_dataframe(rows)

    @property
    def _is_time_series_project(self):
        from datarobot.models import Project
        project = Project.get(self.project_id)
        return project.use_time_series

    @staticmethod
    def _build_timeseries_dataframe(rows):
        columns = (
            'row_id',
            'partition_id',
            'prediction',
            'timestamp',
            'forecast_point',
            'forecast_distance',
            'series_id',
        )
        data = (
            (
                row.row_id,
                row.partition_id,
                row.prediction,
                row.timestamp,
                row.forecast_point,
                row.forecast_distance,
                row.series_id,
            ) for row in rows
        )
        return pd.DataFrame.from_records(data, columns=columns)

    @staticmethod
    def _build_regression_dataframe(rows):
        columns = ('row_id', 'partition_id', 'prediction')
        data = ((row.row_id, row.partition_id, row.prediction) for row in rows)
        return pd.DataFrame.from_records(data, columns=columns)

    @staticmethod
    def _build_classification_dataframe(rows, columns):
        data_list = []
        for row in rows:
            data_row = (row.row_id, row.partition_id, row.prediction)
            data_row += tuple(
                prediction['value'] for prediction in row.prediction_values
            )
            data_list.append(data_row)

        return pd.DataFrame(data_list, columns=columns)

    @staticmethod
    def _get_classification_columns(class_labels, class_prefix):
        columns = ('row_id', 'partition_id', 'prediction')
        columns += tuple(
            '{}{}'.format(class_prefix, label)
            for label in class_labels
        )
        return columns

    def download_to_csv(self, filename, encoding='utf-8'):
        """
        Save training prediction rows into CSV file.

        Parameters
        ----------
        filename : str or file object
            path or file object to save training prediction rows
        encoding : string, optional
            A string representing the encoding to use in the output file, defaults to
            'utf-8'
        """
        df = self.get_all_as_dataframe()
        df.to_csv(
            path_or_buf=filename,
            header=True,
            index=False,
            encoding=encoding,
        )



from datetime import datetime
import responses

from datarobot import DataStore

from datarobot.utils import parse_time


@responses.activate
def test_list(data_stores_endpoint, data_stores_list_server_resp):
    responses.add(responses.GET, data_stores_endpoint, json=data_stores_list_server_resp)

    data_stores = DataStore.list()

    for data_store, server_payload in zip(data_stores, data_stores_list_server_resp['data']):
        assert data_store.id == server_payload['id']
        assert data_store.type == server_payload['type']
        assert data_store.canonical_name == server_payload['canonicalName']
        assert data_store.creator == server_payload['creator']
        assert data_store.updated == parse_time(server_payload['updated'])
        assert data_store.params.driver_id == server_payload['params']['driverId']
        assert data_store.params.jdbc_url == server_payload['params']['jdbcUrl']


@responses.activate
def test_get(data_stores_endpoint, postgresql_data_store_server_resp):
    data_store_id = postgresql_data_store_server_resp['id']
    responses.add(
        responses.GET,
        '{}{}/'.format(data_stores_endpoint, data_store_id),
        json=postgresql_data_store_server_resp
    )

    data_store = DataStore.get(data_store_id)

    assert data_store.id == postgresql_data_store_server_resp['id']
    assert data_store.type == postgresql_data_store_server_resp['type']
    assert data_store.canonical_name == postgresql_data_store_server_resp['canonicalName']
    assert data_store.creator == postgresql_data_store_server_resp['creator']
    assert data_store.updated == parse_time(postgresql_data_store_server_resp['updated'])
    assert data_store.params.driver_id == postgresql_data_store_server_resp['params']['driverId']
    assert data_store.params.jdbc_url == postgresql_data_store_server_resp['params']['jdbcUrl']


@responses.activate
def test_future_proof(data_stores_endpoint, postgresql_data_store_server_resp):
    aug_res = dict(postgresql_data_store_server_resp, newest='NEW_KEY')
    aug_res['params']['newKey'] = 'hihihi'
    data_store_id = aug_res['id']
    responses.add(responses.GET, '{}{}/'.format(data_stores_endpoint, data_store_id), json=aug_res)

    data_store = DataStore.get(data_store_id)

    assert data_store.id == data_store_id


@responses.activate
def test_create(data_stores_endpoint, postgresql_data_store_server_resp):
    responses.add(responses.POST, data_stores_endpoint, json=postgresql_data_store_server_resp)

    data_store = DataStore.create(
        data_store_type=postgresql_data_store_server_resp['type'],
        canonical_name=postgresql_data_store_server_resp['canonicalName'],
        driver_id=postgresql_data_store_server_resp['params']['driverId'],
        jdbc_url=postgresql_data_store_server_resp['params']['jdbcUrl'],
    )

    assert data_store.id == postgresql_data_store_server_resp['id']
    assert data_store.type == postgresql_data_store_server_resp['type']
    assert data_store.canonical_name == postgresql_data_store_server_resp['canonicalName']
    assert data_store.creator == postgresql_data_store_server_resp['creator']
    assert data_store.updated == parse_time(postgresql_data_store_server_resp['updated'])
    assert data_store.params.driver_id == postgresql_data_store_server_resp['params']['driverId']
    assert data_store.params.jdbc_url == postgresql_data_store_server_resp['params']['jdbcUrl']


@responses.activate
def test_update(data_stores_endpoint, postgresql_data_store_server_resp):
    data_store_id = postgresql_data_store_server_resp['id']
    responses.add(
        responses.GET,
        '{}{}/'.format(data_stores_endpoint, data_store_id),
        json=postgresql_data_store_server_resp)
    updated_data_store_server_resp = dict(
        postgresql_data_store_server_resp,
        canonicalName='updated_canonical_name',
        params={
            'driverId': 'updated_driver_id',
            'jdbcUrl': 'updated_jdbc_url'
        }
    )
    responses.add(
        responses.PATCH,
        '{}{}/'.format(data_stores_endpoint, data_store_id),
        json=updated_data_store_server_resp)

    data_store = DataStore.get(data_store_id)

    data_store.update(
        canonical_name=updated_data_store_server_resp['canonicalName'],
        driver_id=updated_data_store_server_resp['params']['driverId'],
        jdbc_url=updated_data_store_server_resp['params']['jdbcUrl'],
    )

    assert data_store.id == updated_data_store_server_resp['id']
    assert data_store.type == updated_data_store_server_resp['type']
    assert data_store.canonical_name == updated_data_store_server_resp['canonicalName']
    assert data_store.creator == updated_data_store_server_resp['creator']
    assert data_store.updated == parse_time(postgresql_data_store_server_resp['updated'])
    assert data_store.params.driver_id == updated_data_store_server_resp['params']['driverId']
    assert data_store.params.jdbc_url == updated_data_store_server_resp['params']['jdbcUrl']


@responses.activate
def test_unable_to_change_type(data_stores_endpoint, postgresql_data_store_server_resp):
    data_store_id = postgresql_data_store_server_resp['id']
    responses.add(
        responses.GET,
        '{}{}/'.format(data_stores_endpoint, data_store_id),
        json=postgresql_data_store_server_resp)
    responses.add(
        responses.PATCH,
        '{}{}/'.format(data_stores_endpoint, data_store_id),
        json=postgresql_data_store_server_resp)

    data_store = DataStore.get(data_store_id)
    data_store._type = 'new_type'
    data_store.update()
    new_data_store = DataStore.get(data_store_id)

    assert new_data_store.type == postgresql_data_store_server_resp['type']


@responses.activate
def test_unable_to_change_creator(data_stores_endpoint, postgresql_data_store_server_resp):
    data_store_id = postgresql_data_store_server_resp['id']
    responses.add(
        responses.GET,
        '{}{}/'.format(data_stores_endpoint, data_store_id),
        json=postgresql_data_store_server_resp)
    responses.add(
        responses.PATCH,
        '{}{}/'.format(data_stores_endpoint, data_store_id),
        json=postgresql_data_store_server_resp)

    data_store = DataStore.get(data_store_id)
    data_store._creator = 'new_creator'
    data_store.update()
    new_data_store = DataStore.get(data_store_id)

    assert new_data_store.creator == postgresql_data_store_server_resp['creator']


@responses.activate
def test_unable_to_change_updated_time(data_stores_endpoint, postgresql_data_store_server_resp):
    data_store_id = postgresql_data_store_server_resp['id']
    responses.add(
        responses.GET,
        '{}{}/'.format(data_stores_endpoint, data_store_id),
        json=postgresql_data_store_server_resp)
    responses.add(
        responses.PATCH,
        '{}{}/'.format(data_stores_endpoint, data_store_id),
        json=postgresql_data_store_server_resp)

    data_store = DataStore.get(data_store_id)
    data_store._updated = datetime.now()
    data_store.update()
    new_data_store = DataStore.get(data_store_id)

    assert new_data_store.updated == parse_time(postgresql_data_store_server_resp['updated'])


@responses.activate
def test_delete(data_stores_endpoint, postgresql_data_store_server_resp):
    data_store_id = postgresql_data_store_server_resp['id']
    responses.add(
        responses.GET,
        '{}{}/'.format(data_stores_endpoint, data_store_id),
        json=postgresql_data_store_server_resp
    )
    delete_url = '{}{}/'.format(data_stores_endpoint, data_store_id)
    responses.add(responses.DELETE, delete_url)

    data_store = DataStore.get(data_store_id)
    data_store.delete()

    assert responses.calls[1].request.method == responses.DELETE
    assert responses.calls[1].request.url == delete_url


@responses.activate
def test_connection_test(data_stores_endpoint, postgresql_data_store_server_resp):
    data_store_id = postgresql_data_store_server_resp['id']
    responses.add(
        responses.GET,
        '{}{}/'.format(data_stores_endpoint, data_store_id),
        json=postgresql_data_store_server_resp
    )
    responses.add(
        responses.POST,
        '{}{}/test/'.format(data_stores_endpoint, data_store_id),
        json={'message': 'Connection successful'})

    data_store = DataStore.get(data_store_id)
    resp = data_store.test(username='my_username', password='my_password')

    assert resp['message'] == 'Connection successful'


@responses.activate
def test_schemas_list(data_stores_endpoint, postgresql_data_store_server_resp,
                      postgresql_schemas_list_server_resp):
    data_store_id = postgresql_data_store_server_resp['id']
    responses.add(
        responses.GET,
        '{}{}/'.format(data_stores_endpoint, data_store_id),
        json=postgresql_data_store_server_resp
    )
    responses.add(
        responses.POST,
        '{}{}/schemas/'.format(data_stores_endpoint, data_store_id),
        json=postgresql_schemas_list_server_resp
    )

    data_store = DataStore.get(data_store_id)
    resp = data_store.schemas(username='my_username', password='my_password')

    assert resp == postgresql_schemas_list_server_resp


@responses.activate
def test_tables_list(
        data_stores_endpoint, postgresql_data_store_server_resp, postgresql_tables_list_server_resp
):
    data_store_id = postgresql_data_store_server_resp['id']
    responses.add(
        responses.GET,
        '{}{}/'.format(data_stores_endpoint, data_store_id),
        json=postgresql_data_store_server_resp
    )
    responses.add(
        responses.POST,
        '{}{}/tables/'.format(data_stores_endpoint, data_store_id),
        json=postgresql_tables_list_server_resp
    )

    data_store = DataStore.get(data_store_id)
    resp = data_store.tables(username='my_username', password='my_password', schema='demo')

    assert resp == postgresql_tables_list_server_resp



import json

import pytest
import responses
from numpy import isclose

from datarobot import Model
from datarobot.models import ModelParameters


@pytest.fixture
def model_parameters_data():
    return {
        'parameters': [
            {'name': u'Intercept', 'value': 1},
            {'name': u'Link function', 'value': u'logit'}
        ],
        'derivedFeatures': [
            {
                'coefficient': -0.015,
                'originalFeature': u'A1Cresult',
                'derivedFeature': u'A1Cresult->7',
                'type': u'CAT',
                'transformations': [{'name': u'One-hot', 'value': u"'>7'"}]
            },
            {
                'coefficient': 0.014,
                'originalFeature': u'A1Cresult',
                'derivedFeature': u'A1Cresult->8',
                'type': u'CAT',
                'transformations': [{'name': u'One-hot', 'value': u"'>8'"}]
            },
            {
                'coefficient': -0.133,
                'originalFeature': u'A1Cresult',
                'derivedFeature': u'A1Cresult-Norm',
                'type': u'CAT',
                'transformations': [{'name': u'One-hot', 'value': u"'Norm'"}]
            },
            {
                'coefficient': 0.40552547397855254,
                'type': u'NUM',
                'derivedFeature': u'STANDARDIZED_revol_util',
                'originalFeature': u'revol_util',
                'transformations': [
                    {'name': u'Missing imputation', 'value': 47.6},
                    {'name': u'Standardize', 'value': u'(47.781034375,27.955074618)'}]}
        ]}


@pytest.fixture
def freq_sev_model_parameters_data(model_parameters_data):
    freq_sev_data = model_parameters_data.copy()
    for derived_feature in freq_sev_data['derivedFeatures']:
        derived_feature['stageCoefficients'] = [{
            'stage': 'frequency',
            'coefficient': 1.0
        }, {
            'stage': 'severity',
            'coefficient': 2.0
        }]
    return freq_sev_data


@pytest.fixture
def broken_model_parameters_data(model_parameters_data):
    broken_data = model_parameters_data.copy()
    bad_data = {'some_broken_field': 'some_broken_value'}
    broken_data['parameters'][0].update(bad_data)
    broken_data['derivedFeatures'][0].update(bad_data)
    broken_data.update(bad_data)
    return broken_data


@pytest.fixture
def model_parameters_url(project_id, model_id):
    return 'https://host_name.com/projects/{}/models/{}/parameters/'.format(
        project_id,
        model_id
    )


@responses.activate
@pytest.mark.usefixtures('client')
def test_instantiation(model_parameters_url, model_parameters_data, project_id, model_id):
    responses.add(
        responses.GET,
        model_parameters_url,
        status=200,
        content_type='application/json',
        body=json.dumps(model_parameters_data)
    )

    mp = ModelParameters.get(project_id, model_id)

    params = mp.parameters
    expected_params = model_parameters_data['parameters']
    for i in range(len(params)):
        assert params[i]['name'] == expected_params[i]['name']
        assert params[i]['value'] == expected_params[i]['value']

    derived = mp.derived_features
    expected_derived = model_parameters_data['derivedFeatures']
    for i in range(len(derived)):
        assert isclose(derived[i]['coefficient'], expected_derived[i]['coefficient'])
        assert derived[i]['original_feature'] == expected_derived[i]['originalFeature']
        assert derived[i]['derived_feature'] == expected_derived[i]['derivedFeature']
        assert derived[i]['type'] == expected_derived[i]['type']
        assert derived[i]['transformations'] == expected_derived[i]['transformations']


@responses.activate
@pytest.mark.usefixtures('client')
def test_freq_sev_parameters(model_parameters_url, freq_sev_model_parameters_data,
                             project_id, model_id):
    responses.add(
        responses.GET,
        model_parameters_url,
        status=200,
        content_type='application/json',
        body=json.dumps(freq_sev_model_parameters_data)
    )

    mp = ModelParameters.get(project_id, model_id)

    derived_features = mp.derived_features
    expected_derived_features = freq_sev_model_parameters_data['derivedFeatures']
    for i in range(len(derived_features)):
        assert len(derived_features[i]['stage_coefficients']) == 2
        for j in range(len(derived_features[i]['stage_coefficients'])):
            assert isclose(derived_features[i]['stage_coefficients'][j]['coefficient'],
                           expected_derived_features[i]['stageCoefficients'][j]['coefficient'])


@responses.activate
@pytest.mark.usefixtures('client')
def test_future_proof(model_parameters_url, broken_model_parameters_data, project_id, model_id):
    responses.add(
        responses.GET,
        model_parameters_url,
        status=200,
        content_type='application/json',
        body=json.dumps(broken_model_parameters_data)
    )

    mp = ModelParameters.get(project_id, model_id)

    assert len(mp.parameters) == 2
    for i in range(len(mp.parameters)):
        assert 'some_broken_field' not in mp.parameters[i]

    assert len(mp.derived_features) == 4
    for i in range(len(mp.derived_features)):
        assert 'some_broken_field' not in mp.derived_features[i]


@responses.activate
@pytest.mark.usefixtures('client')
def test_get_parameters(model_parameters_url, model_parameters_data, project_id, model_id):
    responses.add(
        responses.GET,
        model_parameters_url,
        status=200,
        content_type='application/json',
        body=json.dumps(model_parameters_data)
    )

    mp = ModelParameters.get(project_id, model_id)
    model = Model(id=model_id, project_id=project_id)
    assert mp.parameters == model.get_parameters().parameters



# coding: utf-8
import json
import responses
import pytest

from datarobot import PredictionDataset, Project
from datarobot.utils import parse_time, datetime_to_string
from tests.test_helpers import fixture_file_path
from tests.utils import assert_equal_py2, assert_equal_py3


@pytest.fixture
def list_dataset_json():
    return """
    {"count": 1,
     "next": null,
     "previous": null,
     "data": [
         {
             "id": "ds-id",
             "projectId": "p-id",
             "name": "My PredictionDataset",
             "created": "2016-02-16T12:00:00.123456Z",
             "numRows": 100,
             "numColumns": 20,
             "forecastPoint": null
         }
     ]
    }
    """


@pytest.fixture
def dataset_json(list_dataset_json):
    data = json.loads(list_dataset_json)
    return json.dumps(data['data'][0])


@pytest.fixture
def dataset_with_forecast_point_server_data(dataset_json):
    dataset = dict(json.loads(dataset_json))
    dataset['forecastPoint'] = '2017-01-01T15:00:00Z'
    return dataset


@pytest.fixture
def dataset_with_start_end_server_data(dataset_json):
    dataset = dict(json.loads(dataset_json))
    dataset['predictionsStartDate'] = '2017-01-01T15:00:00Z'
    dataset['predictionsEndDate'] = '2017-06-01T15:00:00Z'
    return dataset


dataset_response = {'id': 'ds-id',
                    'projectId': 'p-id',
                    'name': 'My PredictionDataset',
                    'created': '2016-02-16T12:00:00.123456Z',
                    'numRows': 100,
                    'numColumns': 20,
                    'forecastPoint': None}

datasets_list_response = {'count': 1,
                          'next': None,
                          'previous': None,
                          'data': [dataset_response]}


def prep_successful_upload_response(upload_type, body):
    responses.add(responses.POST,
                  'https://host_name.com/projects/p-id/predictionDatasets/{}/'.format(upload_type),
                  body='',
                  status=202,
                  content_type='application/json',
                  adding_headers={'Location': 'https://host_name.com/status/status-id/'}
                  )
    responses.add(responses.GET,
                  'https://host_name.com/status/status-id/',
                  status=303,
                  body='',
                  content_type='application/json',
                  adding_headers={
                      'Location': 'https://host_name.com/projects/p-id/predictionDatasets/ds-id/'}
                  )
    responses.add(responses.GET,
                  'https://host_name.com/projects/p-id/predictionDatasets/ds-id/',
                  status=200,
                  body=body,
                  content_type='application/json')


def assert_matches_expected_dataset(actual_dataset):
    assert actual_dataset.project_id == dataset_response['projectId']
    assert actual_dataset.id == dataset_response['id']
    assert actual_dataset.name == dataset_response['name']
    assert actual_dataset.created == dataset_response['created']
    assert actual_dataset.num_rows == dataset_response['numRows']
    assert actual_dataset.num_columns == dataset_response['numColumns']
    assert actual_dataset.forecast_point is None


def test_future_proof():
    PredictionDataset.from_server_data(dict(dataset_response, new_key='future'))


@responses.activate
@pytest.mark.usefixtures('client')
def test_list_datasets(list_dataset_json):
    responses.add(responses.GET,
                  'https://host_name.com/projects/p-id/predictionDatasets/',
                  body=list_dataset_json)
    project = Project('p-id')
    datasets = project.get_datasets()
    assert len(datasets) == 1
    dataset = datasets[0]
    assert_matches_expected_dataset(dataset)


@responses.activate
@pytest.mark.usefixtures('client')
def test_retrieve_dataset():
    responses.add(responses.GET,
                  'https://host_name.com/projects/p-id/predictionDatasets/ds-id/',
                  body=json.dumps(dataset_response))
    dataset = PredictionDataset.get('p-id', 'ds-id')
    assert_matches_expected_dataset(dataset)


@responses.activate
@pytest.mark.usefixtures('client')
def test_dataset_repr():
    responses.add(responses.GET,
                  'https://host_name.com/projects/p-id/predictionDatasets/ds-id/',
                  body=json.dumps(dataset_response))
    dataset = PredictionDataset.get('p-id', 'ds-id')
    assert_equal_py2(repr(dataset), "PredictionDataset(u\'My PredictionDataset\')")
    assert_equal_py3(repr(dataset), "PredictionDataset('My PredictionDataset')")


def test_dataset_repr_unicode():
    dataset = PredictionDataset('projectId', 'datasetId', name=u'赤木',
                                created='2016-02-16T12:00:00.123456Z',
                                num_rows=100, num_columns=20)
    assert_equal_py2(repr(dataset), "PredictionDataset(u'\\u8d64\\u6728')")
    assert_equal_py3(repr(dataset), "PredictionDataset('赤木')")


@responses.activate
@pytest.mark.usefixtures('client')
def test_upload_dataset_from_url(dataset_json):
    prep_successful_upload_response('urlUploads', dataset_json)
    project = Project('p-id')
    file_url = 'https://my_files.com/my_dataset.csv'
    new_dataset = project.upload_dataset(file_url)
    assert_matches_expected_dataset(new_dataset)
    assert responses.calls[0].request.body == json.dumps({'url': file_url})


@responses.activate
@pytest.mark.usefixtures('client')
def test_upload_dataset_from_url_with_forecast(dataset_with_forecast_point_server_data):
    prep_successful_upload_response('urlUploads',
                                    json.dumps(dataset_with_forecast_point_server_data))
    project = Project('p-id')
    file_url = 'https://my_files.com/my_dataset.csv'
    forecast_point = parse_time(dataset_with_forecast_point_server_data['forecastPoint'])
    new_dataset = project.upload_dataset(file_url, forecast_point=forecast_point)

    assert new_dataset.forecast_point == forecast_point
    request = json.loads(responses.calls[0].request.body)
    assert parse_time(request['forecastPoint']) == forecast_point


@responses.activate
@pytest.mark.usefixtures('client')
def test_upload_dataset_from_file(dataset_json):
    prep_successful_upload_response('fileUploads', dataset_json)
    filepath = fixture_file_path('onehundredrows.csv')
    project = Project('p-id')
    new_dataset = project.upload_dataset(filepath)
    assert_matches_expected_dataset(new_dataset)
    assert 'file' in responses.calls[0].request.body.fields


@responses.activate
@pytest.mark.usefixtures('client')
def test_upload_dataset_from_file_with_forecast(dataset_with_forecast_point_server_data):
    prep_successful_upload_response('fileUploads',
                                    json.dumps(dataset_with_forecast_point_server_data))
    filepath = fixture_file_path('onehundredrows.csv')
    project = Project('p-id')
    forecast_point = parse_time(dataset_with_forecast_point_server_data['forecastPoint'])
    new_dataset = project.upload_dataset(filepath, forecast_point=forecast_point)

    assert new_dataset.forecast_point == forecast_point
    assert 'forecastPoint' in responses.calls[0].request.body.fields


@responses.activate
@pytest.mark.usefixtures('client')
def test_upload_dataset_from_url_with_start_end(dataset_with_start_end_server_data):
    prep_successful_upload_response('urlUploads',
                                    json.dumps(dataset_with_start_end_server_data))
    project = Project('p-id')
    file_url = 'https://my_files.com/my_dataset.csv'
    predictions_start_date = parse_time(dataset_with_start_end_server_data['predictionsStartDate'])
    predictions_end_date = parse_time(dataset_with_start_end_server_data['predictionsEndDate'])

    new_dataset = project.upload_dataset(file_url,
                                         predictions_start_date=predictions_start_date,
                                         predictions_end_date=predictions_end_date)
    assert new_dataset.predictions_start_date == predictions_start_date
    assert new_dataset.predictions_end_date == predictions_end_date
    request = json.loads(responses.calls[0].request.body)
    assert parse_time(request['predictionsStartDate']) == predictions_start_date
    assert parse_time(request['predictionsEndDate']) == predictions_end_date


@responses.activate
@pytest.mark.usefixtures('client')
def test_upload_dataset_from_url_fails_when_one_date_provided(dataset_with_start_end_server_data):
    project = Project('p-id')
    file_url = 'https://my_files.com/my_dataset.csv'
    predictions_start_date = parse_time(dataset_with_start_end_server_data['predictionsStartDate'])
    msg = 'Both prediction_start_date and prediction_end_date must be provided at the same time'
    with pytest.raises(ValueError, message=msg):
        project.upload_dataset(file_url, predictions_start_date=predictions_start_date)


@responses.activate
@pytest.mark.usefixtures('client')
def test_upload_dataset_from_url_fails_when_fp_and_dates_provided(
        dataset_with_start_end_server_data):
    project = Project('p-id')
    file_url = 'https://my_files.com/my_dataset.csv'
    dataset_with_start_end_server_data['forecastPoint'] = '2017-01-01T15:00:00Z'
    forecast_point = parse_time(dataset_with_start_end_server_data['forecastPoint'])
    predictions_start_date = parse_time(dataset_with_start_end_server_data['predictionsStartDate'])
    predictions_end_date = parse_time(dataset_with_start_end_server_data['predictionsStartDate'])
    msg = ('forecast_point can not be provided together with predictions_start_date or '
           'predictions_end_date')
    with pytest.raises(ValueError, message=msg):
        project.upload_dataset(file_url, forecast_point=forecast_point,
                               predictions_start_date=predictions_start_date,
                               predictions_end_date=predictions_end_date)


@responses.activate
@pytest.mark.usefixtures('client')
def test_upload_dataset_from_file_with_start_end(dataset_with_start_end_server_data):
    prep_successful_upload_response('fileUploads',
                                    json.dumps(dataset_with_start_end_server_data))
    filepath = fixture_file_path('onehundredrows.csv')
    project = Project('p-id')
    predictions_start_date = parse_time(dataset_with_start_end_server_data['predictionsStartDate'])
    predictions_end_date = parse_time(dataset_with_start_end_server_data['predictionsEndDate'])

    new_dataset = project.upload_dataset(filepath,
                                         predictions_start_date=predictions_start_date,
                                         predictions_end_date=predictions_end_date)
    assert new_dataset.predictions_start_date == predictions_start_date
    assert new_dataset.predictions_end_date == predictions_end_date
    assert 'predictionsStartDate' in responses.calls[0].request.body.fields
    assert 'predictionsEndDate' in responses.calls[0].request.body.fields


@responses.activate
@pytest.mark.usefixtures('client')
def test_upload_dataset_from_file_fails_when_one_date_provided(dataset_with_start_end_server_data):
    filepath = fixture_file_path('onehundredrows.csv')
    project = Project('p-id')
    predictions_start_date = parse_time(dataset_with_start_end_server_data['predictionsStartDate'])
    msg = 'Both prediction_start_date and prediction_end_date must be provided at the same time'
    with pytest.raises(ValueError, message=msg):
        project.upload_dataset(filepath, predictions_start_date=predictions_start_date)


@responses.activate
@pytest.mark.usefixtures('client')
def test_upload_dataset_from_file_fails_when_fp_and_dates_provided(
        dataset_with_start_end_server_data):
    filepath = fixture_file_path('onehundredrows.csv')
    project = Project('p-id')
    dataset_with_start_end_server_data['forecastPoint'] = '2017-01-01T15:00:00Z'
    forecast_point = parse_time(dataset_with_start_end_server_data['forecastPoint'])
    predictions_end_date = parse_time(dataset_with_start_end_server_data['predictionsStartDate'])
    msg = ('forecast_point can not be provided together with predictions_start_date or '
           'predictions_end_date')
    with pytest.raises(ValueError, message=msg):
        project.upload_dataset(filepath, forecast_point=forecast_point,
                               predictions_end_date=predictions_end_date)


@responses.activate
@pytest.mark.usefixtures('client')
def test_upload_dataset_from_data_source(data_sources_endpoint, data_sources_list_server_resp,
                                         airlines10mb_data_source_server_resp, dataset_json):
    data_source_id = airlines10mb_data_source_server_resp['id']
    username = 'my_username'
    password = 'my_password'

    prep_successful_upload_response('dataSourceUploads', dataset_json)
    responses.add(responses.GET, data_sources_endpoint, json=data_sources_list_server_resp)
    project = Project('p-id')

    project.upload_dataset_from_data_source(data_source_id, username=username, password=password)

    assert responses.calls[0].request.method == 'POST'
    request_body = json.loads(responses.calls[0].request.body)
    assert request_body == {
        'dataSourceId': data_source_id,
        'user': username,
        'password': password
    }


@responses.activate
@pytest.mark.usefixtures('client')
def test_upload_dataset_from_data_source_with_forecast(
    data_sources_endpoint, data_sources_list_server_resp, airlines10mb_data_source_server_resp,
    dataset_with_forecast_point_server_data
):
    data_source_id = airlines10mb_data_source_server_resp['id']
    forecast_point = parse_time(dataset_with_forecast_point_server_data['forecastPoint'])
    username = 'my_username'
    password = 'my_password'
    prep_successful_upload_response('dataSourceUploads',
                                    json.dumps(dataset_with_forecast_point_server_data))
    responses.add(responses.GET, data_sources_endpoint, json=data_sources_list_server_resp)
    project = Project('p-id')

    project.upload_dataset_from_data_source(data_source_id, username, password,
                                            forecast_point=forecast_point)

    assert responses.calls[0].request.method == 'POST'
    request_body = json.loads(responses.calls[0].request.body)
    assert request_body == {
        'dataSourceId': data_source_id,
        'forecastPoint': datetime_to_string(forecast_point),
        'user': username,
        'password': password,
    }


@responses.activate
@pytest.mark.usefixtures('client')
def test_delete_dataset():
    responses.add(responses.DELETE,
                  'https://host_name.com/projects/p-id/predictionDatasets/ds-id/',
                  body='')
    dataset = PredictionDataset.from_server_data(dataset_response)
    dataset.delete()
    assert responses.calls[0].request.method == 'DELETE'
    assert len(responses.calls) == 1
